"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import numpy as np
from . import _utils

from .data import component_data
from .leaks import Leak


def init_component_sets(species, phase, leak_sizes=None,
                        n_compressors=0, n_vessels=0, n_valves=5, n_instruments=3, n_joints=35, n_hoses=1, n_pipes=20, n_filters=0,
                        n_flanges=0, n_exchangers=0, n_vaporizers=0, n_arms=0, n_extra1s=0, n_extra2s=0,
                        compressor_ps=None, vessel_ps=None, valve_ps=None, instrument_ps=None, joint_ps=None,
                        hose_ps=None, pipe_ps=None, filter_ps=None, flange_ps=None, exchanger_ps=None,
                        vaporizer_ps=None, arm_ps=None, extra1_ps=None, extra2_ps=None,
                        fuel_component_data=None):
    if fuel_component_data is None:
        fuel_component_data = component_data.h2_gas_params

    if leak_sizes is None:
        leak_sizes = _utils.get_default_leak_sizes()

    if compressor_ps is None:
        compressor_ps = fuel_component_data['compressor']
    if vessel_ps is None:
        vessel_ps = fuel_component_data['vessel']
    if valve_ps is None:
        valve_ps = fuel_component_data['valve']
    if instrument_ps is None:
        instrument_ps = fuel_component_data['instrument']
    if joint_ps is None:
        joint_ps = fuel_component_data['joint']
    if hose_ps is None:
        hose_ps = fuel_component_data['hose']
    if pipe_ps is None:
        pipe_ps = fuel_component_data['pipe']
    if filter_ps is None:
        filter_ps = fuel_component_data['filter']
    if flange_ps is None:
        flange_ps = fuel_component_data['flange']
    if exchanger_ps is None:
        exchanger_ps = fuel_component_data['exchanger']
    if vaporizer_ps is None:
        vaporizer_ps = fuel_component_data['vaporizer']
    if arm_ps is None:
        arm_ps = fuel_component_data['arm']
    if extra1_ps is None:
        extra1_ps = fuel_component_data['extra1']
    if extra2_ps is None:
        extra2_ps = fuel_component_data['extra2']

    component_sets = [
        ComponentSet(category='compressor', num_components=n_compressors, species=species, saturated_phase=phase,
                     leak_frequency_lists=compressor_ps, leak_sizes=leak_sizes),
        ComponentSet(category='vessel', num_components=n_vessels, species=species, saturated_phase=phase,
                     leak_frequency_lists=vessel_ps, leak_sizes=leak_sizes),
        ComponentSet(category='valve', num_components=n_valves, species=species, saturated_phase=phase,
                     leak_frequency_lists=valve_ps, leak_sizes=leak_sizes),
        ComponentSet(category='instrument', num_components=n_instruments, species=species, saturated_phase=phase,
                     leak_frequency_lists=instrument_ps, leak_sizes=leak_sizes),
        ComponentSet(category='joint', num_components=n_joints, species=species, saturated_phase=phase,
                     leak_frequency_lists=joint_ps, leak_sizes=leak_sizes),
        ComponentSet(category='hose', num_components=n_hoses, species=species, saturated_phase=phase,
                     leak_frequency_lists=hose_ps, leak_sizes=leak_sizes),
        ComponentSet(category='pipe', num_components=n_pipes, species=species, saturated_phase=phase,
                     leak_frequency_lists=pipe_ps, leak_sizes=leak_sizes),
        ComponentSet(category='filter', num_components=n_filters, species=species, saturated_phase=phase,
                     leak_frequency_lists=filter_ps, leak_sizes=leak_sizes),
        ComponentSet(category='flange', num_components=n_flanges, species=species, saturated_phase=phase,
                     leak_frequency_lists=flange_ps, leak_sizes=leak_sizes),
        ComponentSet(category='exchanger', num_components=n_exchangers, species=species, saturated_phase=phase,
                     leak_frequency_lists=exchanger_ps, leak_sizes=leak_sizes),
        ComponentSet(category='vaporizer', num_components=n_vaporizers, species=species, saturated_phase=phase,
                     leak_frequency_lists=vaporizer_ps, leak_sizes=leak_sizes),
        ComponentSet(category='arm', num_components=n_arms, species=species, saturated_phase=phase,
                     leak_frequency_lists=arm_ps, leak_sizes=leak_sizes),
        ComponentSet(category='extra1', num_components=n_extra1s, species=species, saturated_phase=phase,
                     leak_frequency_lists=extra1_ps, leak_sizes=leak_sizes),
        ComponentSet(category='extra2', num_components=n_extra2s, species=species, saturated_phase=phase,
                     leak_frequency_lists=extra2_ps, leak_sizes=leak_sizes)
    ]
    return component_sets


class ComponentSet:
    """
    Representation of 1+ components of a specific type or category.
    Has associated LeakSet which defines all leaks for this component type.

    Parameters
    ----------
    category : str
        Name of component type, e.g. vessel

    num_components : int
        Number of components in this category/type

    species : str or dict
        If string, expect one of {'h2', 'ch4', 'c3h8'}.
        If dict, fluid is blend with key of species, value of concentration.


    saturated_phase : None or str
        CoolProp parameter for fuel phase.
        Note that these values may be unintuitive. 'gas' and 'liquid' indicate saturated state and will
        use the liquid leak frequency data.
        Typical h2 gas should have parameter value of None.

    leak_frequency_lists: list or None
        List of lists; Leak frequency parameters for each leak size of format [[mu, sigma]]

    leak_sizes : list or None
        List of floats representing % leakage

    Attributes
    -----------
    category : str
        name of component type, e.g. 'compressor'

    num_components : int
        number of components in analysis

    leaks : dict of {size: LeakSet}

    """
    
    default_leak_sizes = _utils.get_default_leak_sizes()
    
    def __init__(self, category, num_components, species, saturated_phase=None, leak_frequency_lists=None, leak_sizes=None):
        if type(category) != str:
            raise ValueError("Category must be string")
        if num_components < 0:
            raise ValueError("# of components must be 0 or greater")

        if leak_sizes is None:
            leak_sizes = self.default_leak_sizes
        
        leak_sizes.sort()
        if leak_sizes[-1] != 100:
            leak_sizes.append(100)

        self.category = category
        self.num_components = num_components
        self.leaks = {}

        # Get or convert leak frequency data as [{mu, sigma}]
        if leak_frequency_lists is None:
            leak_frequency_lists = component_data.get_component_leak_parameters(self.category,
                                                                                species=species,
                                                                                saturated_phase=saturated_phase)

        # Create leak representations for each size
        for leak_size in leak_sizes:
            leak_freq_pair = self.interpolate_leak_size_dist(leak_size, leak_frequency_lists)
            leak = Leak(leak_size, leak_freq_pair[0], leak_freq_pair[1])
            self.leaks[leak.size] = leak

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "{} ({}) freqs: {}".format(self.category, self.num_components, self.get_leaks_str_simple())
    
    def interpolate_leak_size_dist(self, leak_size, leak_frequency_lists):
        alpha_1 = 1
        alpha_2 = 1
        #predicted_mean = alpha_1 + alpha_2*np.log(leak_size)
        # TODO: should this refer to default leak sizes or provided leak sizes?
        predicted_mean = np.interp(np.log(leak_size), np.log(self.default_leak_sizes), np.array(leak_frequency_lists)[:, 0])
        predicted_sigma = np.interp(np.log(leak_size), np.log(self.default_leak_sizes), np.array(leak_frequency_lists)[:, 1])
        return predicted_mean, predicted_sigma

    def get_leak(self, leak_size):
        """ Retrieve leak object at specified size for this component type. """
        if self.num_components:
            if leak_size in self.leaks.keys():
                return self.leaks[leak_size]
            else:
                raise KeyError('Leak of specified size not found')
        else:
            return None

    def get_leak_frequency(self, leak_size):
        freq = 0
        leak = self.get_leak(leak_size)
        if leak:
            freq = leak.get_leak_freq_mean() * self.num_components
        return freq

    def get_leaks_str_simple(self):
        """ simplified version of above as '[x, y], [x2, y2]... ' """
        leak_list = sorted(self.leaks.values(), key=lambda x: x.size)
        full_str = ''
        for i, leak in enumerate(leak_list):
            full_str += '[{}, {}], '.format(leak.mu, leak.sigma)
        full_str = full_str[:-2]  # remove final ", " and end of list
        return full_str



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import copy
import warnings
import time

import numpy as np
import matplotlib.pyplot as plt
from scipy import integrate, optimize

from ._therm import CoolPropWrapper
from ..utilities.custom_warnings import PhysicsWarning
from ..utilities import misc_utils



class Fluid:
    def __init__(self, species=None, T=None, P=None, rho=None, phase=None, v=0, therm=None):
        """
        class used to describe a fluid (usually a gas)
        two of four (T, P, rho, phase) are needed to fully define the fluid
        
        Parameters
        ----------
        species: string or dictionary
            species (either formula or name - see CoolProp documentation) or dictionary of {species:molefraction}
        T : float
            temperature (K)
        P: float
            pressure (Pa)
        rho: float
            density (kg/m^3)
        v: float
            velocity (m/s)
        phase: {None, 'gas', 'liquid'}
            either 'gas' or 'liquid' if fluid is at the saturated state.
        therm : thermodynamic class
            a thermodynamic class that is used to relate state variables

        """
        if misc_utils.is_parsed_blend_string(species):
            self.species = species

        else:
            if type(species) == str:
                cleaned = misc_utils.parse_fluid_key(species)
                if misc_utils.is_valid_fluid_string(cleaned):
                    self.species = cleaned
                else:
                    raise ValueError(f"'{cleaned}' is not a valid fluid name")

            elif type(species) == dict:
                single_species = len(species) == 1
                # Rebuild dict to parse and validate fluid names
                species = misc_utils.parse_blend_dict_into_coolprop_dict(species)
                self.species = misc_utils.parse_coolprop_dict_into_string(species)
                if single_species:  # e.g. Hydrogen[1.00] -> Hydrogen
                    self.species = self.species.split('[')[0]

        if therm is None:
            therm = CoolPropWrapper(self.species)

        params = therm._init_fluid_params({'D':rho, 'P':P, 'T':T, 'phase':phase})
        
        self.rho, self.P, self.T, self.phase = params['D'], params['P'], params['T'], params['phase']
        self.therm = therm
        self.v = v

    def update(self, T=None, P=None, rho=None, v=None):
        if v != None:
            self.v = v

        if T is not None or P is not None or rho is not None:
            params = self.therm._init_fluid_params({'D':rho, 'P':P, 'T':T, 'phase':None})
            self.rho = params['D']
            self.P = params['P']
            self.T = params['T']
            self.phase = params['phase']

    def __repr__(self):
        return 'Gas\n%s\n  P = %.3f bar\n  T = %0.1f K\n  rho = %.3f kg/m^3)\n  v = %.1f m/s' % (
            30 * '-', self.P * 1e-5, self.T, self.rho, self.v)


class Orifice:
    def __init__(self, d, Cd=1):
        '''
        class used to describe a circular orifice
        
        future versions may be expanded to give effective area for other shapes
        
        Parameters
        ----------
        d - orifice diameter (m)
        Cd - discharge coefficient to account for non-plug flow (always <=1, assumed to be 1 for plug flow)
        
        Contains
        --------
        d - diameter (m)
        Cd- discharge coefficient 
        A - effective area (m^2)
        '''
        self.d, self.Cd, self.A = d, Cd, np.pi / 4 * d ** 2

    def __repr__(self):
        return 'orifice\n%s\ndiameter = %.2f mm\ndischarge coefficient = %.2f' % (30 * '-', self.d * 1e3, self.Cd)

    def mdot(self, fluid):
        '''
        mass flow rate through orifice of a fluid object
        
        Parameters
        ----------
        rho - density (kg/m^3)
        v - velocity (m/s)
        
        Returns
        -------
        mdot - mass flow rate (kg/s)
        '''
        return fluid.rho * fluid.v * self.A * self.Cd

    def flow(self, upstream_fluid, downstream_P = 101325, mdot = None, suppressWarnings = True):
        '''
        Returns the fluid in a flow restriction, for given upstream conditions 
        and downstream pressure.  Isentropic expansion.
        
        Parameters
        ----------
        upstream_fluid - upstream fluid with therm object, as well as P, T, rho, v
        downstream_P - float: downstream pressure (Pa)
        mdot - mass flow rate - only used for unchoked flow
        
        Returns
        -------
        Fluid object containing T, P, rho, v at the throat (orifice)
        '''           
        h0 = upstream_fluid.therm.get_property('H0', P=upstream_fluid.P, D=upstream_fluid.rho, v=upstream_fluid.v)
        if upstream_fluid.v > 0:
            s0 = upstream_fluid.therm.get_property('S', H = h0, D = np.round(upstream_fluid.rho, 12))
        else: #LH2 simulations were giving weird results when calculating entropy from enthalpy
            s0 = upstream_fluid.therm.get_property('S', D = upstream_fluid.rho, T = upstream_fluid.T)
        
        fluid = copy.copy(upstream_fluid)
        if fluid.P < downstream_P:
            raise ValueError('Downstream pressure is higher than upstream pressure.  Nonphysical.')
        if fluid.P == downstream_P:
            if mdot is not None:
                v = mdot/(self.Cd*fluid.rho*self.A)
                fluid.update(v=v)
                fluid._choked = False
                return fluid
            else:
                raise ValueError('Downstream pressure is the same as upstream pressure.  Need to specify mass flow rate (mdot).')

        def negflux(P):
            h, rho = fluid.therm.get_property(['H', 'D'], P=P, S=s0)
            return -rho*np.sqrt(2 * (h0 - h))

        P = optimize.minimize_scalar(negflux, bounds = (downstream_P, upstream_fluid.P), method = 'bounded')['x']
        h, rho = fluid.therm.get_property(['H', 'D'], P = P, S = s0)
        fluid.update(rho = rho, P = P, v = np.sqrt(2 * (h0 - h)))
        if P - downstream_P > .01: 
            if mdot is not None:
                if not suppressWarnings:
                    warnings.warn('Fluid choked. Ignoring mdot specification and using choked flow calculation.', category=PhysicsWarning)
            fluid._choked = True
        else: 
            fluid._choked = False
            if mdot is None:
                if not suppressWarnings:
                    warnings.warn('Fluid unchoked. Provide a mass flow rate.', category=PhysicsWarning)
            else:
                v = mdot/(self.Cd*rho*self.A)
                fluid.update(rho=rho, P=(P*(P - downstream_P > .01) + 
                                         downstream_P*(P - downstream_P <= .01)), v=v)
        return fluid


class Source(object):
    """
    Used to describe a source (tank) that contains a fluid

    Attributes
    ----------
    mass : float
        mass of source (kg)

    """

    def __init__(self, V, fluid):
        '''
        Initializes source based on the volume and the fluid object in the source (tank)
        
        Parameters
        ----------
        V: float
            volume of source (tank) (m^3)
        fluid : Fluid
            fluid object in the source (tank)

        Returns
        -------
        source: object
            object containing .fluid (fluid obejct), .V (volume, m^3), and .m (mass (kg))
        '''
        self.fluid = fluid
        self.V = V
        self.m = self.mass = fluid.rho * V

    @classmethod
    def fromMass(cls, m, fluid):
        '''
        Initilization method based on the mass and the fluid object in the source (tank)
        
        Parameters
        ----------
        m: float
            mass of source (tank) (kg)
        fluid: object
            fluid object in the source (tank)
        
        Returns
        -------
        source: object
            object containing .fluid (fluid obejct), .V (volume, m^3), and .m (mass (kg))
        '''
        V = m / fluid.rho
        return cls(V, fluid)

    @classmethod
    def fromMass_Vol(cls, species, m, V, T=None, P=None):
        '''
        Initilization method based on the mass, volume, and either the temperature or pressure of the fluid 
        in the source (tank).
        
        Parameters
        ----------
        species: string
            species (either formula or name - see CoolProp documentation)
        m: float
            mass of source (tank) (kg)
        V: float
            volume of source (tank) (m^3)
        therm: object
            thermodynamic class used to relate pressure, temperature and density
        T: float (optional)
            temperature (K)
        P: float (optional)
            pressure (Pa)
        
        Returns
        -------
        source: object
            object containing .fluid (fluid object), .V (volume, m^3), and .m (mass (kg)).
        returns none if either underspecified (neither T or P given) or overspecified (both T and P given)
        '''
        rho = m / V
        if T is not None and P is None:
            fluid = Fluid(species=species, rho=rho, T=T)
        elif T is None and P is not None:
            fluid = Fluid(species=species, rho=rho, P=P)
        else:
            return None
        return cls(V, fluid)

    def _blowdown_gov_eqns(self, t, ind_vars, Vol, orifice, heat_flux, ambient_P):
        '''governing equations for energy balance on a tank (https://doi.org/10.1016/j.ijhydene.2011.12.047)
        
        Parameters
        ----------
        t - time (s)
        ind_vars - array of mass (kg), internal energy (J/kg) in tank
        Vol - float, volume of tank (m^3)
        orifice - orifice object
        heat_flux - float, heat flow into tank (W)
        
        Returns
        -------
        [dm_dt, du_dt] = array of [d(mass)/dt (kg/s), d(internal energy)/dt (J/kg-s)]
                       = [-rho_throat*v_throat*A_throat, 1/m*(Q_in + mdot_out*(u - h_out))]
        '''
        therm = self.fluid.therm
        m, U = ind_vars
        m, U = float(m), float(U)
        rho = m / Vol
        T = therm.get_property('T', U=U, D=np.round(rho, 10))
        fluid = copy.copy(self.fluid)
        fluid.update(T=T, rho=rho)
        throat = orifice.flow(fluid, ambient_P)
        h = therm.get_property('H', T=fluid.T, D=fluid.rho)
        dm_dt = -orifice.mdot(throat)
        du_dt = 1 / m * (heat_flux + (h - U) * dm_dt)
        return np.array([dm_dt, du_dt])
   
    def empty(self, orifice, ambient_P = 101325,
              heat_flux = 0, nmax = 1000, 
              m_empty = 1e-6, p_empty_percent = 0.01,
              t_empty = None,
              timeout = 10):
        '''
        integrates the governing equations for an energy balance on a tank 
        
        Parameters
        ----------
        orifice - orifice object through which the source is emptying
        ambient_P - ambient pressure into which leak occurs (Pa)
        heat_flux - Heat flow (W) into tank.  Assumed to be 0 (adiabatic)
        nmax - maximum number of iterations
        m_empty - mass when considered empty (kg)
        p_empty_percent - percent of ambient pressure when considered empty
        timeout - computer run time (minutes) until you give up on the integration
        t_empty - if given, empties for a set amount of time (sec)
        
        Returns
        -------
        tuple of (mdot, fluid_list, t, solution_array) =
                 (list of mass flow rates (kg/s), list of fluid objects at each time step, 
                  array of times (s), 2D array of [mass, internal energy] at each time step)
        '''
        start_time = time.time()
        therm = self.fluid.therm
        m0 = self.m
        volume = self.V
        A_orifice = orifice.A * orifice.Cd
        T0, rho0, P = self.fluid.T, self.fluid.rho, self.fluid.P
        u0 = therm.get_property('U', T=T0, D=rho0)
        r = integrate.ode(self._blowdown_gov_eqns).set_integrator('dopri5')
        r.set_initial_value([m0, u0]).set_f_params(volume, orifice, heat_flux, ambient_P)
        throat = orifice.flow(self.fluid, ambient_P)
        if t_empty is None:
            mdot0 = orifice.mdot(throat)
            dt = m0 / mdot0
        else:
            dt = t_empty
        times, fluid_list, mdot, sol = [], [], [], []

        def solout(t, y):
            if len(times) > 0:
                if times[-1] == t:
                    return
            rho = y[0] / volume
            T = therm.get_property('T', U=y[1], D=np.round(rho, 10))
            fluid = copy.copy(self.fluid)
            fluid.update(T=T, rho=rho)
            throat = orifice.flow(fluid, ambient_P)
            fluid_list.append(fluid)
            mdot.append(orifice.mdot(throat))
            times.append(t)
            sol.append([y[0], y[1]])

        r.set_solout(solout)
        nsteps = 0
        while (True):
            nsteps += 1
            try:
                r.integrate(r.t + dt)
            except:
                for i in range(100):
                    try:
                        # need to back up to last successful integration point
                        j = np.argmin(np.abs(np.array(times) - r.t))
                        times = times[:j + 1]
                        fluid_list = fluid_list[:j + 1]
                        mdot = mdot[:j + 1]
                        sol = sol[:j + 1]
                        dt /= 10
                        try:
                            r.integrate(r.t + dt)
                            break
                        except:
                            pass
                    except:
                        print(i, 'unable to advance past time %.5f s with remaining mass of %.3f g, P = %.1f Pa' %
                              (r.t, r.y[0] * 1000, fluid_list[-1].P))
                        return mdot, fluid_list, times, np.array(sol).T
            if time.time() - start_time > timeout*60:
                print(f'timeout of {timeout} min reached')
                break
            if nsteps > nmax:
                print('max steps reached')
                break
            if not r.successful():
                break
            if t_empty is not None:
                if r.t >= t_empty:
                    break
            elif (r.y[0] < m_empty or fluid_list[-1].P < (1 + p_empty_percent / 100) * ambient_P):
                break
        self.mdot, self.fluid_list, self.ts, self.sol = mdot, fluid_list, times, np.array(sol).T 
        return mdot, fluid_list, times, np.array(sol).T
    
    def plot_time_to_empty(self):
        try:
            x = self.sol[0]
        except:
            return None
        fig, axs = plt.subplots(4, 1, sharex=True, squeeze=True, figsize=(4, 7))
        axs[0].plot(self.ts, self.sol[0])
        axs[0].set_ylabel('Mass [kg]')
        axs[1].plot(self.ts, np.array([f.P for f in self.fluid_list]) * 1e-5)
        axs[1].set_ylabel('Pressure [bar]')
        axs[2].plot(self.ts, self.mdot)
        axs[2].set_ylabel('Flow Rate [kg/s]')
        axs[3].plot(self.ts, [f.T for f in self.fluid_list])
        axs[3].set_ylabel('Temperature [K]')
        axs[3].set_xlabel('Time [s]')
        [a.minorticks_on() for a in axs]
        [a.grid(which='major', color='k', dashes=(2, 2), alpha=.5) for a in axs]
        [a.grid(which='minor', color='k', alpha=.1) for a in axs]
        return fig

class Enclosure:
    '''
    Enclosure used in the overpressure modeling
    '''

    def __init__(self, H, A, H_release, ceiling_vent, floor_vent, Xwall=np.inf):
        '''
        Describes the enclosure
        
        Parameters
        ----------
        H : encosure height (m)
        A : area of floor and ceiling (m^2) 
        H_release : height of release (m)
        ceiling_vent : vent class containing vent information for ceiling vent
        floor_vent : vent class containing vent information for floor vent
        Xwall : perpendicular from jet to wall (m)
        '''
        self.H, self.A, self.ceiling_vent, self.floor_vent = H, A, ceiling_vent, floor_vent
        self.H_release, self.Xwall = H_release, Xwall
        self.V = H * A


class Vent:
    '''
    Vent used in overpressure modeling
    '''

    def __init__(self, A, H, Cd=1, vol_flow_rate=0):
        '''
        Describes the vent
        
        Parameters
        ----------
        A : vent cross-sectional area (m^2)
        H : vent height from floor (m)
        Cd: discharge coefficient of vent
        vol_flow_rate: volumetric flow rate through the vent (m^3/s)
        '''
        self.A, self.H, self.Cd, self.vol_flow_rate = A, H, Cd, vol_flow_rate
        self.Qw = Cd * vol_flow_rate / np.sqrt(2)  # See Lowesmith et al IJHE 2009



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import copy
import os

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
from mpl_toolkits.axes_grid1 import ImageGrid

from ._fuel_props import FuelProperties


def plot_sliced_contour(contours, xlims, ylims, zlims, nx, ny, nz,
                        distance_func, value_func, slice_xyz,
                        colorbar_label,
                        origin_lines=None,
                        title=None,
                        savefig=False,
                        directory=os.getcwd(),
                        filename='sliced_contour_plot.png',
                        *args, **kwargs):
    '''
    Get grid location values for use in sliced contour plot

    Parameters
    ----------
    contours : list-like
        contour levels to show on plot;
        only first value is used to determine furthest contour
    xlims : tuple/list or None
        x-limits for plot
        (if None, will be determined based on furthest contour)
    ylims : tuple/list or None
        y-limits for plot
        (if None, will be determined based on furthest contour)
    zlims : tuple/list or None
        z-limits for plot
        (if None, will be determined based on furthest contour)
    nx : int
        number of points to solve for in the x-direction
    ny : int
        number of points to solve for in the y-direction
    nz : int
        number of points to solve for in the z-direction
    distance_func : callable
        function/method that will calculate the distance to furthest contour;
        must accept an effect value as the first argument argument,
        a keyword argument of direction,
        followed by whatever other arguments are used
    value_func : callable
        function/method that will calculate values for contour plot;
        must accept list-likes of x-values, y-values, and z-values
        as the first three arguments,
        followed by whatever other arguments are used
    slice_xyz : list of 3 values
        The (x, y, z) coordinate of where the 3 slices are located
    colorbar_label : string
        label for colorbar
    origin_lines : list of 3 list-like (optional)
        if provided, this is a list of values to plot black lines
        [x-values, y-values, z-values]
    title : string (optional)
        if provided, title shown on plot
    savefig : Boolean (optional)
        determines if figure file is saved
    directory : string (optional)
        directory in which to save file
    filename : string (optional)
        file name to write
    *args : positional arguments (optional)
        if provided, passed to distance_func
    **kwargs : keyword arguments, optional
        if provided, passed to distance_func

    Returns
    -------
    fig_or_filepath: matplotlib.pyplot.Figure object or string
        If savefig is True, returns filename of the corresponding plot.
        If savefig is false, returns fig object.
    '''
    furthest_contour = contours[0] * 1000  # convert from kilo
    default_scaling = 1.1  # default scaling is 10% past furthest contour

    if xlims is None:
        pos_x_distance_to_contour = distance_func(furthest_contour,
                                                  direction='x',
                                                  *args, **kwargs)
        neg_x_distance_to_contour = distance_func(furthest_contour,
                                                  direction='x',
                                                  negative_direction=True,
                                                  *args, **kwargs)
        padded_pos_x_distance = default_scaling * pos_x_distance_to_contour
        padded_neg_x_distance = default_scaling * neg_x_distance_to_contour
        dx = (padded_pos_x_distance - padded_neg_x_distance) / nx
        x0 = slice(padded_neg_x_distance, padded_pos_x_distance, dx)
    else:
        dx = (xlims[1] - xlims[0]) / nx
        x0 = slice(xlims[0], xlims[1], dx)

    if ylims is None:
        y_distance_to_contour = distance_func(furthest_contour,
                                              direction='y',
                                              *args, **kwargs)
        padded_y_distance = default_scaling * y_distance_to_contour
        dy = padded_y_distance / ny
        y0 = slice(0, padded_y_distance, dy)
    else:
        dy = (ylims[1] - ylims[0]) / ny
        y0 = slice(ylims[0], ylims[1], dy)

    if zlims is None:
        z_distance_to_contour = distance_func(furthest_contour,
                                              direction='z',
                                              *args, **kwargs)
        padded_z_distance = default_scaling * z_distance_to_contour
        dz = 2 * padded_z_distance / nz
        z0 = slice(-padded_z_distance, padded_z_distance, dz)
    else:
        dz = (zlims[1] - zlims[0]) / nz
        z0 = slice(zlims[0], zlims[1], dz)

    x_z, y_z = np.mgrid[x0, y0]
    x_y, z_y = np.mgrid[x0, z0]
    y_x, z_x = np.mgrid[y0, z0]

    fxy = value_func(x_z, y_z, slice_xyz[2] * np.ones_like(x_z), *args, **kwargs)
    fxz = value_func(x_y, slice_xyz[1] * np.ones_like(x_y), z_y, *args, **kwargs)
    fzy = value_func(slice_xyz[0] * np.ones_like(z_x), y_x, z_x, *args, **kwargs)
    fxy = fxy / 1000  # convert to kilo
    fxz = fxz / 1000
    fzy = fzy / 1000

    fig = plt.figure(figsize=(8, 4.5))
    # Not sure if this will always be right; tight_layout incompatible with ImageGrid
    fig.subplots_adjust(top=0.967,
                        bottom=0.378)
    grid = ImageGrid(fig, 111,  # similar to subplot(111)
                     nrows_ncols=(2, 2),  # creates 2x2 grid of axes
                     axes_pad=0.1,  # pad between axes in inches.
                     label_mode="L",
                     cbar_mode='edge',
                     cbar_location='bottom',
                     cbar_size='10%',
                     cbar_pad=-1.5)
    ax_xy, ax_zy, ax_xz, ax_cb = grid[0], grid[1], grid[2], grid[3]

    for ax in [ax_xy, ax_zy, ax_xz]:
        ax.cax.set_visible(False)
        ax.minorticks_on()
        ax.grid(alpha=0.2, color='k')
        ax.grid(which='minor', alpha=0.1, color='k')
        ax.set_aspect(1)
    ax_zy.axis['bottom'].toggle(all=True)
    ax_cb.set_frame_on(False)
    ax_cb.set_axis_off()
    ax_cb.cax.set_visible(True)

    # ClrMap = copy.copy(plt.cm.get_cmap('RdYlGn_r'))
    ClrMap = copy.copy(mpl.colormaps['RdYlGn_r'])
    ClrMap.set_under('white')

    ax_xy.contourf(x_z, y_z, fxy, cmap=ClrMap,
                    levels=contours, extend='both')
    if origin_lines is not None:
        ax_xy.plot(origin_lines[0], origin_lines[1],
                   color='k', linewidth=3, solid_capstyle='round')
    ax_xy.set_ylabel('Height (y) [m]')
    ax_xy.annotate(f'z = {slice_xyz[2]:.2f} m', xy=(0.02, 0.98),
                   xycoords='axes fraction', va='top', color='k')

    ax_xz.contourf(x_y, z_y, fxz, cmap=ClrMap,
                    levels=contours, extend='both')
    if origin_lines is not None:
        ax_xz.plot(origin_lines[0], origin_lines[2],
                   color='k', linewidth=3, solid_capstyle='round')
    ax_xz.set_xlabel('Horizontal Distance (x) [m]')
    ax_xz.set_ylabel('Perpendicular Distance (z) [m]')
    ax_xz.annotate(f'y = {slice_xyz[1]:.2f} m', xy=(0.02, 0.98),
                   xycoords='axes fraction', va='top', color='k')

    im = ax_zy.contourf(z_x, y_x, fzy, cmap=ClrMap,
                        levels=contours, extend='both')
    if origin_lines is not None:
        ax_zy.plot(origin_lines[2], origin_lines[1],
                   color='k', linewidth=3, solid_capstyle='round')
    ax_zy.set_xlabel('Perpendicular Distance (z) [m]')
    ax_zy.annotate(f'x = {slice_xyz[0]:.2f} m', xy=(0.02, 0.98),
                   xycoords='axes fraction', va='top', color='k')

    cb = plt.colorbar(im, cax=ax_cb.cax, orientation='horizontal', extendfrac='auto')
    cb.set_label(colorbar_label)

    if title is not None:
        fig.suptitle(title)

    if savefig:
        filepath = os.path.join(directory, filename)
        fig.savefig(filepath, bbox_inches='tight')
        plt.close(fig)
        return filepath
    else:
        return fig


def plot_contour(data_type, jet_or_flame, mcolors='w',
                 xlims=None, ylims=None, vlims=None,
                 contour_levels=None, add_colorbar=True, plot_color=None,
                 plot_title=None, x_label=None, y_label=None,
                 fig_params=None, plot_params=None, subplot_params=None, fig=None, ax=None):
    '''
    Creates contour plot.
    Currently handles temperature, mole fraction, mass fraction, and velocity curves. Specify type using the 'data_type' parameter.

    Parameters
    ----------
    data_type: str
        Specifies the type of plot (controls labels and default colors). Options are 'temperature', 'mole', 'mass', and 'velocity'
    jet_or_flame: object
        Jet or Flame object
    mcolors: color or list of colors, optional
        Color(s) of marked contour levels
    xlims: tuple(float or None) or None, optional
        X-axis boundaries of (xmin, xmax) for contour plot
    ylims: tuple(float or None) or None, optional
        Y-axis boundaries of (ymin, ymax) for contour plot
    vlims: tuple(float or None) or None, optional
        Value (z-axis) boundaries of (vmin, vmax) for contour plot
    contour_levels: int or list, optional
        Contour levels to mark and label
    add_colorbar: boolean, optional
        Add a colorbar to the plot. Defaults to True
    plot_color: str, optional
        Matplotlib Colormap name to use for the contour plot. Default colors are set by the different types of plot
    plot_title: str, optional
        Title text for the plot
    x_label: str, optional
        X-axis label for the plot
    y_label: str, optional
        Y-axis label for the plot
    fig_params: dict or None, optional
        Dictionary of figure parameters (e.g. figsize)
    plot_params: dict or None, optional
        Dictionary of contour plot parameters (e.g. colors)
    subplot_params: dict or None, optional
        Dictionary of subplots_adjust parameters (e.g. top)
    fig: matplotlib Figure, optional
        Figure on which to make the plot, used if figure already exists before this function is called
    ax: matplotlib Axes, optional
        Axes on which to make the plot, used if figure already exists 

    Returns
    -------
    fig: matplotlib Figure
        Figure containing the contour plot
    '''
    # Set params to dicts
    if fig_params is None:
        fig_params = {}
    if plot_params is None:
        plot_params = {}
    if subplot_params is None:
        subplot_params = {}

    # Make figure and axis if not specified
    if ax is None and fig is None:
        fig, ax = plt.subplots(**fig_params)
        plt.subplots_adjust(**subplot_params)
    elif ax is None and fig is not None:
        ax = fig.subplots(**fig_params)
        plt.subplots_adjust(**subplot_params)
    else:
        fig = ax.figure

    # Format plot
    ax.set_aspect('equal')
    if xlims is not None:
        ax.set_xlim(*xlims)
    if ylims is not None:
        ax.set_ylim(*ylims)        
    if plot_title is not None:
        ax.set_title(plot_title)

    if x_label is not None:
        ax.set_xlabel(x_label)
    else:
        ax.set_xlabel('x (m)')

    if y_label is not None:
        ax.set_ylabel(y_label)
    else:
        ax.set_ylabel('y (m)')

    # Contour plot parameters based on type selection
    if data_type == "temperature":
        x, y, v = jet_or_flame._contourdata
        cb_label = 'Temperature (K)'
        if not plot_color:
            plot_color = 'plasma'
        
    elif data_type == "mole":
        x, y, v, __, __, __ = jet_or_flame._contourdata
        cb_label = 'Mole Fraction'
        if not plot_color:
            plot_color = 'viridis'

    elif data_type == "mass":
        x, y, __, v, __, __ = jet_or_flame._contourdata
        cb_label = 'Mass Fraction'
        if not plot_color:
            plot_color = 'viridis'
        
    elif data_type == "velocity":
        x, y, __, __, v, __ = jet_or_flame._contourdata
        cb_label = 'Velocity (m/s)'    
        if not plot_color:
            plot_color = 'viridis'

    else:
        raise ValueError('Invalid datatype')

    # Adjust v-limits for filled contour colors
    if vlims is not None:
        vmin = vlims[0]
        vmax = vlims[1]
        if vmin is None and vmax is not None:
            if np.amax(v) > vmax: 
                extend = 'max'
            else: 
                extend = 'neither'
            vmin = np.amin(v)
        elif vmax is None and vmin is not None:
            if np.amin(v) < vmin: 
                extend = 'min'
            else: 
                extend = 'neither'
            vmax = np.amax(v)
        else:
            if np.amax(v) > vmax and np.amin(v) < vmin: 
                extend = 'both'
            elif np.amax(v) > vmax and np.amin(v) >= vmin: 
                extend = 'max'
            elif np.amax(v) <= vmax and np.amin(v) < vmin: 
                extend = 'min'
            else: 
                extend = 'neither'
    else:
        vmin = np.amin(v)
        vmax = np.amax(v)
        extend = 'neither'

    filled_levels = np.linspace(vmin, vmax, 101)

    # Filled contour and background
    plt.set_cmap(plot_color)
    if hasattr(mpl, 'colormaps'):
        ax.set_facecolor(mpl.colormaps[plot_color](0))
    else:
        ax.set_facecolor(mpl.cm.get_cmap(plot_color)(0))

    filled_contour = ax.contourf(x, y, v, levels=filled_levels, vmin=vmin, vmax=vmax, extend=extend, **plot_params)
    
    # Add specific line contours 
    if contour_levels is not None:
        contour_levels = [contour_levels] if type(contour_levels) in [int, float] else contour_levels
        
        # Check for values outside of range
        if data_type == "mole":
            contour_levels = [FuelProperties(jet_or_flame.fluid.species).LFL if value == 'LFL' else value for value in contour_levels]
            contour_levels = [FuelProperties(jet_or_flame.fluid.species).UFL if value == 'UFL' else value for value in contour_levels]

            for contour in contour_levels:
                if contour <= 0 or contour >= 1:
                    error_msg = (f'Mole fraction contour values must be >0 and <1 (current value: {contour})')
                    raise ValueError(error_msg)

        # Draw contour lines
        CS = ax.contour(x, y, v, levels=np.sort(contour_levels), colors=mcolors, linewidths=1.5, **plot_params)

        # Draw contour line labels
        if xlims is not None and ylims is not None:
            manual = [np.interp(m, jet_or_flame.X_cl[::-1], 
                                jet_or_flame.S[::-1]) * np.array([np.cos(jet_or_flame.theta[0]), 
                                                                    np.sin(jet_or_flame.theta[0])])*.7 for m in contour_levels]
        else:
            manual = False

        ax.clabel(CS, levels=CS.levels, inline=True, manual=manual, inline_spacing=2)
    
    # Add color bar
    if add_colorbar:
        cb_kwargs = {}
        cb_label_kwargs = {}
        
        # Automatically set colorbar orientation
        if xlims:
            width = np.abs(xlims[1] - xlims[0])
        else:
            lims = ax.get_xlim()
            width = np.abs(lims[1] - lims[0])

        if ylims:
            height = np.abs(ylims[1] - ylims[0])
        else:
            lims = ax.get_ylim()
            height = np.abs(lims[1] - lims[0])

        if width <= height:
            cb_kwargs = {'orientation':'vertical'}
            cb_label_kwargs = {'rotation':-90, 'va':'bottom'}
        else:
            cb_kwargs = {'orientation':'horizontal'}

        if vmin and vmax:
            colorbar = plt.colorbar(filled_contour, ticks=MaxNLocator().tick_values(vmin, vmax), **cb_kwargs)
        else:
            colorbar = plt.colorbar(filled_contour, **cb_kwargs)

        colorbar.set_label(cb_label, **cb_label_kwargs)
    
    ax.set_aspect(1)
    fig.tight_layout()

    return fig



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.

"""

import os
import unittest
import json

import numpy as np
from scipy.constants import liter, bar, psi
from scipy import interpolate

from hyram.phys import Source, Fluid, Orifice
from hyram.utilities import misc_utils

import utils

"""
NOTE: if running from IDE like pycharm, make sure cwd is hyram/ and not hyram/tests.
"""

# Flags to enable command line output, pyplots, and text file output
VERBOSE = False
CREATE_PLOTS = False
CREATE_OUTPUT = False

# Absolute paths to input/output data
DATA_LOC = os.path.join(os.path.dirname(__file__), 'data')
OUTPUT_LOC = os.path.join(misc_utils.get_temp_folder(), 'validation-blowdown')
OUTPUT_FILE = os.path.join(OUTPUT_LOC, 'Validation.txt')
LIMITS_FILE = os.path.join(DATA_LOC, 'Limits-blowdown.json')


class Test_Ekoto_2012(unittest.TestCase):
    """
    Single test for flow rate over time, data from Ekoto et al. 2012 Figure 3
    """
    def setUp(self):
        data = utils.read_csv(os.path.join(DATA_LOC, 'ekoto-2012-fig3.csv'), header=2)

        with open(LIMITS_FILE) as f:
            self.limits = json.load(f)[__class__.__name__]

        # Set up parameters
        temp = 297  # K...assumed initial tank temp is ambient air temp. Table 1 of 'Exp. investigation of hydrogen release' by Ekoto
        pressure = 13450000 # Pa...Table 1 of above source
        orifice_diameter = 0.00356  # m...Table 1 of above source
        discharge_coeff = 0.75 #...Table 1 of above source
        tank_volume = 0.00363  # m^3...Table 1 of above source
        rel_species = 'H2'

        fluid = Fluid(species=rel_species, P=pressure, T=temp)
        orifice = Orifice(orifice_diameter, discharge_coeff)
        source = Source(tank_volume, fluid)

        # Calculated values
        self.calc_flowrate, _, self.calc_time, _ = source.empty(orifice)

        # Experimental values
        exp_time = np.ravel(data['x3'])
        self.exp_flowrate = np.ravel(data['y3'])

        # Interpolate to calculated data
        interp = interpolate.interp1d(exp_time, self.exp_flowrate, fill_value='extrapolate')
        self.exp_flowrate = interp(self.calc_time)

    def test_flowrate(self):
        title = "Ekoto et al. 2012 - Figure 3, Flow Rate v. Time"
        error_limits = self.limits[title]
        error = utils.get_error(exp_vals=self.exp_flowrate,
                                calc_vals=self.calc_flowrate,
                                error_limits=error_limits,
                                units='kg/s',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=self.calc_time,
                               exp_y=self.exp_flowrate,
                               calc_y=self.calc_flowrate,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Time (s)',
                               ylabel='Flow Rate (kg/s)')


class Test_Schefer_2007(unittest.TestCase):
    """
    Single test for pressure over time, data from Schefer et al. 2007 Figure 4
    """
    def setUp(self):
        data = utils.read_csv(os.path.join(DATA_LOC, 'schefer-2007-fig4.csv'), header=2)

        with open(LIMITS_FILE) as f:
            self.limits = json.load(f)[__class__.__name__]

        # Set up variables
        temp = 290  # K
        pressure = 4.31e7 # Pa
        orifice_diameter = 0.00508  # m
        discharge_coeff = 1
        tank_volume = 1234 * liter  # m^3
        rel_species = 'H2'

        fluid = Fluid(species=rel_species, P=pressure, T=temp)
        orifice = Orifice(orifice_diameter, discharge_coeff)
        source = Source(tank_volume, fluid)
        _, fluid_list, times, _ = source.empty(orifice)

        # Calculated values
        self.calc_time = np.array(times)
        self.calc_pressure = np.array([fluid.P for fluid in fluid_list]) / psi

        x_data = data['x1'] + data['x2']
        y_data = data['y1'] + data['y2']
        sorted_data = sorted(zip(x_data, y_data))

        # Experimental values
        exp_time = []
        self.exp_pressure = []

        def add_data(x, y):
            exp_time.append(x)
            self.exp_pressure.append(y)

        [add_data(x,y) for x, y in sorted_data if x not in exp_time]

        interp = interpolate.interp1d(exp_time, self.exp_pressure, fill_value='extrapolate')
        self.exp_pressure = interp(self.calc_time)

    def test_pressure(self):
        title = "Schefer et al. 2007 - Figure 4, Pressure v. Time"
        error_limits = self.limits[title]
        error = utils.get_error(exp_vals=self.exp_pressure,
                                calc_vals=self.calc_pressure,
                                error_limits=error_limits,
                                units='PSIG',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=self.calc_time,
                               exp_y=self.exp_pressure,
                               calc_y=self.calc_pressure,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Time (s)',
                               ylabel='Pressure (PSIG)')


class Test_Proust_2011(unittest.TestCase):
    """
    Tests for pressure and temperature over time, data from Proust et al. 2011 Figure 4
    """
    def setUp(self):
        data = utils.read_csv(os.path.join(DATA_LOC, 'proust-2011-fig4.csv'), header=1)

        with open(LIMITS_FILE) as f:
            self.limits = json.load(f)[__class__.__name__]

        # Set up variables
        temp = 315.15  # K
        pressure = 90 * 1000000 # Pa
        orifice_diameter = 0.002  # m
        tank_volume = 25 * liter  # m^3
        rel_species = 'H2'

        fluid = Fluid(species=rel_species, P=pressure, T=temp)
        orifice = Orifice(orifice_diameter)
        source = Source(tank_volume, fluid)
        _, fluid_list, times, _ = source.empty(orifice, heat_flux = 10)

        # Only use first 60 seconds
        times = [t for t in times if t < 60]
        fluid_list = fluid_list[:len(times)]

        # Calculated values
        self.calc_time = np.array(times)
        self.calc_pressure = np.array([fluid.P for fluid in fluid_list]) / bar
        self.calc_temp = np.array([fluid.T for fluid in fluid_list]) - 273.15

        # Experimental values
        time_pressure = data['x3']
        time_temp = data['x4']
        self.exp_pressure = data['y3']
        self.exp_temp = data['y4']

        # Interpolate values
        interp = interpolate.interp1d(time_pressure, self.exp_pressure, fill_value='extrapolate')
        self.exp_pressure = interp(self.calc_time)
        interp = interpolate.interp1d(time_temp, self.exp_temp, fill_value='extrapolate')
        self.exp_temp = interp(self.calc_time)

    def test_pressure(self):
        title = "Proust et al. 2011 - Figure 4, Pressure v. Time"
        error_limits = self.limits[title]
        error = utils.get_error(exp_vals=self.exp_pressure,
                                calc_vals=self.calc_pressure,
                                error_limits=error_limits,
                                units='bar',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=self.calc_time,
                               exp_y=self.exp_pressure,
                               calc_y=self.calc_pressure,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Time (s)',
                               ylabel='Pressure (bar)')

    def test_temperature(self):
        title = "Proust et al. 2011 - Figure 4, Temperature v. Time"
        error_limits = self.limits[title]
        error = utils.get_error(exp_vals=self.exp_temp,
                                calc_vals=self.calc_temp,
                                error_limits=error_limits,
                                units='C',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=self.calc_time,
                               exp_y=self.exp_temp,
                               calc_y=self.calc_temp,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Time (s)',
                               ylabel='Temperature (C)')


class Test_Schefer_2006(unittest.TestCase):
    """
    Single test for flow rate over time, data from Schefer et al. 2006 Figure 3b
    """
    def setUp(self):
        data = utils.read_csv(os.path.join(DATA_LOC, 'schefer-2006-fig3b.csv'), header=2)

        with open(LIMITS_FILE) as f:
            self.limits = json.load(f)[__class__.__name__]

        # Set up variables
        temp = 315.15  # K
        pressure = 1.5513e7 # Pa
        d_orifice = 0.003175  # m
        tank_vol = 0.098  # m^3
        rel_species = 'H2'

        fluid = Fluid(species=rel_species, P=pressure, T=temp)
        orifice = Orifice(d_orifice)
        source = Source(tank_vol, fluid)

        # Calculated values
        flowrate, _, self.calc_time, _ = source.empty(orifice)
        self.calc_flowrate = np.array(flowrate) * 1000

        # Experimental values
        exp_time = data['x1']
        self.exp_flowrate = data['y1']

        interp = interpolate.interp1d(exp_time, self.exp_flowrate, fill_value='extrapolate')
        self.exp_flowrate = interp(self.calc_time)

    def test_flowrate(self):
        title = "Schefer et al. 2006 - Figure 3b, Flow Rate v. Time"
        error_limits = self.limits[title]
        error = utils.get_error(exp_vals=self.exp_flowrate,
                                calc_vals=self.calc_flowrate,
                                error_limits=error_limits,
                                units='g/s',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=self.calc_time,
                               exp_y=self.exp_flowrate,
                               calc_y=self.calc_flowrate,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Time (s)',
                               ylabel='Mass Flow Rate (g/s)')


if __name__ == "__main__":
    if os.path.exists(OUTPUT_FILE):
        os.remove(OUTPUT_FILE)
    unittest.main()



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""
import numpy as np


def get_distance_to_effect(value, from_point, direction, effect_func, *args,
                           max_distance=500, interpolation_points=10000,
                           negative_direction=False, **kwargs):
    """
    Calculates the distance from some starting point to a physical effect value of interest

    Parameters
    ----------
    value: float or array-like
        effect value of interest
    from_point: array-like of length 3
        (x, y, z) location of reference point
        from which distance will be measured
    direction: 'x', 'y', or 'z'
        axis along which distance will be calculated
    effect_func: callable
        function/method that will calculate the effect values
        must accept (x_values, y_values, z_values) as arguments
        followed by whatever other arguments are used
    *args: positional arguments (optional)
        if provided, passed to effect_func
    max_distance: float (optional)
        maximum distance from which to calculate distance to effect
        (default value is 500)
    interpolation_points: int (optional)
        number of distance-points at which to calculate effects
        and use for interpolation
        (default value is 10,000)
    negative_direction : Boolean (optional)
        whether or not to look in the negative direction instead of positive
        (default is False)
    **kwargs: keyword arguments, optional
        if provided, passed to effect_func

    Returns
    -------
    distance: float or array-like
        distance to effect value of interest
        calculated from the from_point
        along the direction axis specified
    """
    # TODO: Should this be changed to a solver instead of interpolation? Does 10,000 calculations
    if negative_direction:
        max_distance = -1 * max_distance
    distances = np.linspace(max_distance, 0, interpolation_points)
    if direction == 'x':
        x_values = from_point[0] + distances
        y_values = from_point[1] * np.ones_like(distances)
        z_values = from_point[2] * np.ones_like(distances)
    elif direction == 'y':
        x_values = from_point[0] * np.ones_like(distances)
        y_values = from_point[1] + distances
        z_values = from_point[2] * np.ones_like(distances)
    elif direction == 'z':
        x_values = from_point[0] * np.ones_like(distances)
        y_values = from_point[1] * np.ones_like(distances)
        z_values = from_point[2] + distances
    else:
        raise ValueError(f"Direction ('{direction}') must be 'x', 'y', or 'z'")
    effects = effect_func(x_values, y_values, z_values, *args, **kwargs)
    # effect values must be monotonically increasing
    index_max_value = np.argmax(effects)
    distance = np.interp(value, effects[:index_max_value], distances[:index_max_value])
    return distance



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import copy

import matplotlib.pyplot as plt
import numpy as np
from scipy import integrate, optimize
import scipy.constants as const

from ._fuel_props import FuelProperties
from ._notional_nozzle import NotionalNozzle
from ._plots import plot_contour

########################################################################
# TODO: untested for alternate fuels
########################################################################


class DevelopingFlow:
    def __init__(self, fluid, orifice, ambient, mdot=None,
                 theta0=0, x0=0, y0=0,
                 lam=1.16, betaA=0.28,
                 nn_conserve_momentum=True, nn_T='solve_energy', 
                 T_establish_min=-1,
                 suppressWarnings=False, verbose=False):
        '''
        Engineering correlations to calculate the Gaussian profile boundary conditions for
        the Jet and Flame models
        '''
        self.fluid = fluid
        self.ambient = ambient
        self._lam = lam
        self._betaA = betaA
        self.verbose = verbose
        self._theta0, self.x0, self.y0 = theta0, x0, y0
        self.T_establish_min = T_establish_min
        self._nn_T = nn_T
        self._nn_conserve_momentum = nn_conserve_momentum
        S0 = 0 # S always starts at 0. x and y may start somewhere else.
        Y0 = 1 # pure fluid
        
        # Orifice flow
        self._orifice = orifice
        if self.verbose:
            print('solving for orifice flow... ', end='')
        self.fluid_orifice = self.orifice.flow(fluid, ambient.P, mdot, suppressWarnings) # plug node at orifice exit
        self.mass_flow_rate = self.orifice.mdot(self.fluid_orifice)
        if self.verbose:
            print('done')
        self.d0 = orifice.d
        self.orifice_node = PlugNode(orifice.d*np.sqrt(orifice.Cd), self.fluid_orifice.v, self.fluid_orifice.rho,# diameter scaled to account for Cd
                                     1, self.fluid_orifice.T, theta0, x0, y0, S0)

        # Underexpanded jet (if needed: gets fluid to atmospheric pressure)
        self.fluid_exp, self.orifice_exp = self._expand(orifice, ambient, nn_T, nn_conserve_momentum)
 
        # Initial entrainment and heating (if needed: warms fluid to good T for thermodynamics)
        self.expanded_plug_node = self._dev_plug(self.fluid_exp, self.orifice_exp, ambient, Y0, theta0, x0, y0, S0, 
                                                 T_establish_min, betaA)

        # Develop into established Gaussian profile from plug flow
        self.initial_node = self.expanded_plug_node.establish(ambient, self.fluid_exp, lam)
    
    @property
    def lam(self):
        return self._lam

    @lam.setter
    def lam(self, val):
        self._lam = val
        self.initial_node = self.expanded_plug_node.establish(self.ambient, self.fluid_exp, val)

    @property
    def betaA(self):
        return self._betaA

    @betaA.setter
    def betaA(self, val):
        self.expanded_plug_node = self._dev_plug(self.fluid_exp, self.orifice_exp, self.ambient,
                                                 1, self.theta0, self.x0, self.y0, 0,
                                                 self.T_establish_min, val)
        self.initial_node = self.expanded_plug_node.establish(self.ambient, self.fluid_exp, self.lam)
        self._betaA = val

    @property
    def theta0(self):
        return self._theta0

    @theta0.setter
    def theta0(self, val):
        self.orifice_node.theta = val
        S_out = self.expanded_plug_node.S
        x0, y0 = self.orifice_node.x, self.orifice_node.y
        self.expanded_plug_node.x = x0 + S_out*np.cos(val)
        self.expanded_plug_node.y = y0 + S_out*np.sin(val)
        self.expanded_plug_node.theta = val
        SE = 6.2*self.expanded_plug_node.d
        self.initial_node.x = self.expanded_plug_node.x + SE*np.cos(val)
        self.initial_node.y = self.expanded_plug_node.y + SE*np.sin(val)
        self.initial_node.S = self.expanded_plug_node.S + SE
        self.initial_node.theta = val
        self._theta0 = val

    @property
    def orifice(self):
        return self._orifice

    #todo: could probably improve orifice.setter - can likely redo _expand with less overhead - notional nozzle models may need to be modified
    @orifice.setter
    def orifice(self, new_orifice):
        if self.fluid_orifice._choked: #choked - P, rho, T, v remain constant at orifice (fluid_orifice is constant)
            self.mass_flow_rate = new_orifice.mdot(self.fluid_orifice)
        else: #unchoked - mass flow rate through orifice remains constant
            self.fluid_orifice.v = self.mass_flow_rate/(self.fluid_orifice.rho*new_orifice.Cd*new_orifice.A)
        self._orifice = new_orifice
        self.d0 = new_orifice.d
        self.orifice_node = PlugNode(self.d0*np.sqrt(new_orifice.Cd), self.fluid_orifice.v, self.fluid_orifice.rho,# diameter scaled to account for Cd
                                     1, self.fluid_orifice.T, self.theta0, self.x0, self.y0, self.S0)
        # Underexpanded jet (if needed: gets fluid to atmospheric pressure)
        self.fluid_exp, self.orifice_exp = self._expand(new_orifice, self.ambient, self._nn_T, self._nn_conserve_momentum)

        # Initial entrainment and heating (if needed: warms fluid to good T for thermodynamics)
        self.expanded_plug_node = self._dev_plug(self.fluid_exp, self.orifice_exp, self.ambient,
                                                 1, self.theta0, self.x0, self.y0, 0,
                                                 self.T_establish_min, self.betaA)

        # Develop into established Gaussian profile from plug flow
        self.initial_node = self.expanded_plug_node.establish(self.ambient, self.fluid_exp, self.lam)


    def _expand(self, orifice, ambient, nn_T, nn_conserve_momentum):
        '''
        expands an underexpanded jet, if needed
        '''
        if self.fluid_orifice.P > ambient.P: # use notional nozzle model
            if self.verbose:
                print('solving for notional nozzle... ', end='')
            nn = NotionalNozzle(self.fluid_orifice, orifice, ambient)
            g, o = nn.calculate(nn_T, nn_conserve_momentum)
            if self.verbose:
                print('done.')
        else:
            g, o = self.fluid_orifice, orifice
        return g, o

    def _dev_plug(self, fluid, orifice, ambient, Y0, theta0, x0, y0, S0, 
                  T_establish_min, betaA):
        '''
        zone of initial entrainment and heating (if needed)
        conservation of energy and mass while air being entrained
        '''
        if fluid.T > T_establish_min:
            plug_node = PlugNode(orifice.d, fluid.v, fluid.rho, Y0, fluid.T, theta0, x0, y0, S0)
        else:
            if self.verbose:
                print('solving for zone of initial entrainment and heating... ', end='')
            air_out = copy.copy(ambient)
            air_out.update(T = T_establish_min, P = air_out.P) 
            fluid_out = copy.copy(fluid)
            fluid_out.update(T = T_establish_min, P = fluid_out.P)
            mdot_in = orifice.mdot(fluid)
            h_in = fluid.therm.h(T = fluid.T, D = fluid.rho) + fluid.v**2/2 # assumes mdot_in is pure
            h_air_in = ambient.therm.h(T = ambient.T, P = ambient.P)
            h_air_out = air_out.therm.h(T = air_out.T, D = air_out.rho)
            h_fluid_out = fluid_out.therm.h(T = fluid_out.T, D = fluid_out.rho)
            def errH(Y):
                h_out = (1-Y)*h_air_out + Y*h_fluid_out
                rho_out = 1/((1-Y)/air_out.rho + Y/fluid_out.rho)
                mdot_out = mdot_in + (1-Y)/Y*mdot_in # assumes mdot_in is pure
                v_out = fluid.v*mdot_in/mdot_out # need to think about whether this is right
                h_out = h_out + v_out**2/2
                return mdot_in*h_in + (mdot_out - mdot_in)*h_air_in - mdot_out*h_out
            Y = optimize.brentq(errH, 1.e-6, 1)
            rho_out = 1/((1-Y)/air_out.rho + Y/fluid_out.rho)
            mdot_out = mdot_in + (1-Y)/Y*mdot_in
            v_out = fluid.v*mdot_in/mdot_out
            E = betaA*np.sqrt(orifice.mdot(fluid)*fluid.v/ambient.rho) 
            S_out = (1-Y)*(mdot_out - mdot_in)/E/ambient.rho
            d_out = np.sqrt(mdot_out/(rho_out*v_out)*4/np.pi)
            plug_node = PlugNode(d_out, v_out, rho_out, Y, fluid_out.T, theta0, 
                                 x0 + S_out*np.cos(theta0), y0 + S_out*np.sin(theta0), S0 + S_out)
            if self.verbose:
                print('done.')
        return plug_node

class PlugNode:
    def __init__(self, d, v, rho, Y, T, theta, x = 0, y = 0, S = 0):
        '''Plug flow node
        
        Parameters
        ----------
        d: float, diameter (m)
        v: float, velocity (m/s)
        rho: float, density (kg/m^3)
        Y: float, mass fraction fuel
        theta: float, release angle from horizontal (rad)
        x: float, x position (m),
        y: float, y position (m),
        S: float, distance along streamline (m)
        '''
        self.d, self.v, self.rho, self.Y = d, v, rho, Y
        self.theta, self.x, self.y, self.S = theta, x, y, S
        self.T = T
    
    def Froude(self, rho_amb):
        '''Froude number'''
        return self.v/np.sqrt(const.g*self.d*abs(rho_amb - self.rho)/self.rho) #ESH - 7/23/20 - added abs to work with negatively buoyant jets
    
    def establish(self, ambient, fluid, lam):
        '''
        returns the gaussian node after flow establishment
        '''
        V_clE = self.v
        BE = self.d/np.sqrt(2*(2*lam**2+1)/(lam**2*self.rho/ambient.rho + lam**2 + 1))
        Y_clE = self.Y*(lam**2+1.)/(2.*lam**2)
        h_amb, cp_ambient = ambient.therm.get_property(['H', 'C'], T = ambient.T, P = ambient.P)

        cp_gas = fluid.therm.get_property('C', T = (self.T + ambient.T)/2., P = ambient.P)
        MW_clE = 1 / (Y_clE / fluid.therm.MW + (1 - Y_clE) / ambient.therm.MW)
        h_amb = ambient.T*cp_ambient
        cp_s3 = self.Y * cp_gas + (1 - self.Y) * cp_ambient # this is just cp_gas - using self.Y screws it up
        
        T_s3 = self.T
        # TODO: Use ThermoProps to calculate enthalpy
        h_s3 = cp_s3 * T_s3
        h_clE = h_amb + (lam**2+1)/(2*lam**2)*(h_s3 - h_amb)
        cp_clE = Y_clE * cp_gas + (1 - Y_clE) * cp_ambient
        T_clE = h_clE / cp_clE

        rho_clE = ambient.P*MW_clE/(const.R*T_clE)
        #######################################################################################
        # Note: correlation below is for Fr**2 > 40, which is pretty much always true for these 
        # gases - could add Froude number correlation if desired
        #######################################################################################
        SE = 6.2*self.d
        xE = self.x + SE*np.cos(self.theta)
        yE = self.y + SE*np.sin(self.theta)
        SE = self.S + SE
        thetaE = self.theta
        return GaussianNode(BE, V_clE, rho_clE, Y_clE, thetaE, xE, yE, SE)

class GaussianNode:
    def __init__(self, B, v_cl, rho_cl, Y_cl, theta, x = 0, y = 0, S = 0):
        '''
        Describes the properties of a node 
        
        Parameters
        ----------
        B: float
            Gaussian halfwidth (m)
        v_cl : float
            velocity at centerline (m/s)
        rho_cl : float
            density at centerline (kg/m^3)
        Y_cl : float
            mass fraction at centerline (kg/kg)
        theta: float
            angle of jet (rad, 0 is horizontal)
        x: float
            horizontal position of node (m)
        y: float
            vertical position of node (m)
        S : float (optional)
            length along jet(m)
        '''
        self.B, self.v_cl, self.rho_cl, self.Y_cl = B, v_cl, rho_cl, Y_cl
        self.S, self.theta, self.x, self.y = S, theta, x, y
    
    @property
    def conditions(self):
        return np.array([self.v_cl, self.B, self.rho_cl, self.Y_cl, self.theta, self.x, self.y])    

    def entrainment(self, Emom, rho_amb, alpha_buoy, alpha):
        FrL = self.v_cl**2*self.rho_cl/(const.g*self.B*abs(rho_amb-self.rho_cl)) # ESH: added absolute value 07/23/20 - for negatively buoyant jets, this was a negative number - might need justification
        E_buoy = alpha_buoy/FrL*(2*const.pi*self.v_cl*self.B)*np.sin(self.theta)  # m**2/s
        E = Emom + E_buoy
        alphatest = E/(2*const.pi*self.v_cl*self.B)
        if alphatest > alpha:
            E = alpha*2*const.pi*self.B*self.v_cl
        return E        


class Jet:
    def __init__(self, fluid, orifice, ambient, mdot=None,
                 theta0=0, x0=0, y0=0,
                 lam=1.16, betaA=0.28,
                 nn_conserve_momentum=True, nn_T='solve_energy',
                 T_establish_min=-1,
                 Ymin=7e-4, dS=None, Smax=np.inf,
                 max_steps=5000, tol=1e-8,
                 alpha=0.082, Yamb=0, numB=5, numpts=500,
                 developing_flow = None,
                 suppressWarnings=False, verbose=False):
        '''
        Class for solving for a 2D jet. 
        If fluid pressure is <= 2 x ambient pressure, use subsonic initialization (specify mdot).
               
        Parameters
        ----------
        fluid: fluid object
            the fluid (fluid) that is being released
        orifice: orifice object
            the release
        ambient: fluid object
            the fluid into which the release occurs
        mdot: float, optional defaults to None
            mass flow rate (kg/s). Only used if flow is not choked.
        theta0 : float, optional defaults to horizontal
            angle of release, radians (0 is horizontal, pi/2 is vertical)
        x0: float, optional
            horizontal starting location (m)
        y0: float, optional
            vertical starting location (m)     
        lam : float, optional
            Relative spreading ratio of concentration to velocity
        betaA : float, optional
            proportionality constant for air entrainment (see Li et al IJHE 2015)
        nn_conserve_momentum: boolean, optional
            whether notional nozzle model should conserve mass and momentum, or mass only,
            together with nn_T determines which notional nozzle model to use (see below)
        nn_T: string, optional
            either 'solve_energy', 'Tthroat' or specified temperature (stagnation temperature) 
            with nn_conserve_momentum leads to one of the following notional nozzle models:
            YuceilOtugen - conserve_momentum = True, T = 'solve_energy'
            EwanMoodie - conserve_momentum = False, T = 'Tthroat'
            Birch - conserve_momentum = False, T = T0
            Birch2 - conserve_momentum = True, T = T0
            Molkov - conserve_momentum = False, T = 'solve_energy'
        T_establish_min: float, optional
            minimum temperature (K) at the zone of flow establishment, if specified, will implement the 
            zone of initial entrainment and heating if temperature after notional nozzle model is lower
        Ymin: float, optional
            minimum mass fraction to integrate to (default is about 1 mol%)
        dS: float, optional
            integrator step size, if None, defaults to 500 diameters, solver adds steps in high gradient areas
        Smax: float, optional
            maximum limit of integration, integrator will stop when it reaches Ymin or Smax
        max_steps: float, optional
            maximum steps for integrator
        tol: float, optional
            relative and absolute tolerance for integrator
        alpha: float, optional
            empirical buoyancy induced entrainment constant
        Yamb: float, optional
            mass fraction of fluid in the ambient air
        numB: float, optional
            maximum number of halfwidths (B) considered to be infinity - for integration in energy equations
        numpts: int, optional
            maximum number of points in energy integration (from 0 to numB)
        developing_flow: DevelopingFlow, optional
            specified DevelopingFlow object
        suppressWarnings: boolean, optional
            whether to display warnings about fluid being under-/over-specified in DevelopingFlow object
        verbose: boolean, optional
            whether to include print statements about the model actions
        There are up to 4 engineering models that give initial conditions to an 
        integral model:
        1) flow through the orifice - choked if pressure above critical pressure, assumed
           to be plug flow at volumetric flow rate if not
        2) if choked, underexpanded (notional nozzle)
        3) if choked and < 60K, initial entrainment and heating
        4) flow establishment

        Properties
        ----------
        S : ndarray of floats
            distance (m) along jet?
        '''
        self.verbose = verbose

        if developing_flow is None:
            self.developing_flow = DevelopingFlow(fluid, orifice, ambient, mdot,
                                                  theta0=theta0, x0=x0, y0=y0,
                                                  lam=lam, betaA=betaA,
                                                  nn_conserve_momentum=nn_conserve_momentum, nn_T=nn_T,
                                                  T_establish_min=T_establish_min,
                                                  suppressWarnings=suppressWarnings,
                                                  verbose=verbose)
        else:
            self.developing_flow = developing_flow
        self.initial_node = self.developing_flow.initial_node
        self.mass_flow_rate = self.developing_flow.mass_flow_rate
        
        # Set up some entrainment parameters:
        Fr = self.developing_flow.expanded_plug_node.Froude(ambient.rho)
        self.Fr = Fr
        if Fr < 268:
            self._alpha_buoy = 17.313-0.11665*Fr+(2.0771e-4)*Fr**2 
        else:
            self._alpha_buoy = 0.97
        expanded_plug_node = self.developing_flow.expanded_plug_node
        self._Emom = betaA*np.sqrt(const.pi/4*expanded_plug_node.d**2*
                                   expanded_plug_node.rho*expanded_plug_node.v**2/ambient.rho)
        # some other objects and parameters that I don't want to keep recalculating
        # note: method of calculating heat capacity and enthalpy makes a significant difference in terms
        # of trajectory and density in the calculations...
        self.lam, self.fluid, self.ambient = lam, fluid, ambient
        #####################################################################################################
        # TODO: account for variations in heat capacity as a function of temperature - used when getting rid of
        #       heat capacity in _gov_eqns
        MW_air, MW_fluid = ambient.therm.MW, fluid.therm.MW
        MW_cl0 = MW_air*MW_fluid/(self.initial_node.Y_cl*(MW_air - MW_fluid) + MW_fluid)
        T_cl0 = ambient.P*MW_cl0/(const.R*self.initial_node.rho_cl)
        # TODO: determine if _Cp_fluid should be at ambient T, or T_cl0
        self._Cp_fluid = self.fluid.therm.get_property('C', T = ambient.T, P = ambient.P)
        self._Cp_air, self._h_amb0 = ambient.therm.get_property(['C', 'H'], T = ambient.T, P = ambient.P)

        # Integrate in the zone of established flow
        self.solve(Ymin, dS, Smax, max_steps, tol, alpha, Yamb, numB, numpts)

    @classmethod
    def from_developed_flow(cls, developing_flow,
                            lam=1.16, betaA=0.28,
                            Ymin=7e-4, dS=None, Smax=np.inf,
                            max_steps=5000, tol=1e-8,
                            alpha=0.082, Yamb=0, numB=5, numpts=500,
                            suppressWarnings=False, verbose=False):
        '''
        Initialization of a Jet when the DevelopingFlow calculations have already been made.
        These calculations can be slow for a blend, so if both a Jet and Flame are to be created for the same
        release (e.g. during a QRA), doing the calculations a single time can significantly improve overall
        computation time.'''
        if developing_flow.lam != lam:
            developing_flow.lam = lam
        if developing_flow.betaA != betaA:
            developing_flow.betaA = betaA
        return cls(developing_flow.fluid, developing_flow.orifice, developing_flow.ambient,
                   mdot=None,
                   theta0= 0, x0=0, y0=0,
                   lam=developing_flow.lam, betaA=developing_flow.betaA,
                   nn_conserve_momentum=True, nn_T='solve_energy',
                   T_establish_min=-1,
                   Ymin=Ymin, dS=dS, Smax=Smax,
                   max_steps=max_steps, tol=tol,
                   alpha=alpha, Yamb=Yamb, numB=numB, numpts=numpts,
                   developing_flow = developing_flow,
                   suppressWarnings=suppressWarnings, verbose=verbose)

    def solve(self, Ymin = 7e-4, dS = None, Smax = np.inf, 
              max_steps = 5000, tol = 1e-8,
              alpha = 0.082, Yamb = 0, numB = 5, numpts = 500):
        '''
        solves (integrates) the model equations from the initial node out to limit
        '''
        if self.verbose:
            print('integrating... ', end='')

        if dS is None and Smax == np.inf:
            dS = 500*self.developing_flow.expanded_plug_node.d # somewhat arbitrary - solver will add points anyway 
                                        # may integrate past Ymin 
        elif dS is None:
            dS = Smax

        r = integrate.ode(self._govEqns).set_f_params(alpha, Yamb, numB, numpts)
        r.set_integrator('dopri5', atol = tol, rtol = tol)
        
        T, Y = [], []
        def solout(t, y):
            T.append(t)
            Y.append(np.array(y))
        r.set_solout(solout)
        r.set_initial_value(self.initial_node.conditions, self.initial_node.S)
        
        i = 0
        while r.successful() and r.y[3] > Ymin and i < max_steps and r.t < Smax:
            r.integrate(r.t + dS)
            i += 1
            
        Y = np.array(Y)
        
        for key, val in zip(['V_cl', 'B', 'rho_cl', 'Y_cl', 'theta', 'x', 'y'], Y.T):
            self.__dict__[key] = val
        self.__dict__['S'] = np.array(T)

        MW_fluid, MW_air = self.fluid.therm.MW, self.ambient.therm.MW
        MW_cl  = MW_air*MW_fluid/(self.Y_cl*(MW_air-MW_fluid) + MW_fluid)
        self.X_cl = self.Y_cl*MW_cl/MW_fluid
        self.T_cl = self.ambient.P*MW_cl/(const.R*self.rho_cl)
        
        if self.verbose:
            print('done.')

        return self
    
    def _govEqns(self, S, ind_vars, alpha = 0.082, Yamb = 0, numB = 5, numpts = 500):
        '''
        Governing equations for a plume, written in terms of d/dS of (V_cl, B, rho_cl, Y_cl, 
        theta, x, and y).
        
        A matrix solution to the continuity, x-momentum, y-momentum, species, and energy 
        equations solves for d/dS of the dependent variables V_cl, B, rho_cl, Y_cl,  and Theta.  
        Numerically integrated to infinity = numB * B(S) using numpts discrete points.
        '''
        # break independent variables out of ind_vars, then put them into node_in
        [V_cl, B, rho_cl, Y_cl, theta, x, y] = ind_vars
        node_in = GaussianNode(B, V_cl, rho_cl, Y_cl, theta, S = S)
        
        # pull some parameters out of objects so their definition isn't so long
        rho_amb, MW_air, MW_fluid = self.ambient.rho, self.ambient.therm.MW, self.fluid.therm.MW
        lam = self.lam
        Pamb = self.ambient.P
        h_amb0, Cp_fluid, Cp_air = self._h_amb0, self._Cp_fluid, self._Cp_air
        h_amb0 = Cp_air * self.ambient.T
        E = node_in.entrainment(self._Emom, rho_amb, self._alpha_buoy, alpha = alpha)
        
        # some stuff needed to integrate to infinity (numB*B):
        r = np.append(np.array([0]), np.logspace(-5, np.log10(numB*max(B, 1e-99)), numpts))
        zero = np.zeros_like(r)
        V       = V_cl*np.exp(-(r**2)/(B**2))
        dVdS = np.array([V/V_cl,                                                 #d/dS(V_cl)
                         2*V*r**2/B**3,                                          #d/dS(B)
                         zero,                                                   #d/dS(rho_cl) 
                         zero,                                                   #d/dS(Y_cl)
                         zero])                                                  #d/dS(theta)
        rho     = (rho_cl - rho_amb)*np.exp(-(r**2)/((lam*B)**2))+rho_amb
        Y       = Y_cl*rho_cl/rho*np.exp(-r**2/(lam*B)**2)
        dYdS = np.array([zero,                                                      #d/dS(V_cl)
                         (2*Y**2*rho_amb*r**2*np.exp(r**2/(lam*B)**2)/
                         (lam**2*B**3*Y_cl*rho_cl)),                                #d/dS(B)
                         Y**2*rho_amb*(np.exp(r**2/(lam*B)**2)-1)/(Y_cl*rho_cl**2), #d/dS(rho_cl)
                         Y/Y_cl,                                                    #d/dS(Y_cl)
                         zero])                                                     #d/dS(theta)
        MW      = MW_air*MW_fluid/(Y*(MW_air - MW_fluid) + MW_fluid)
        dMWdS   = (MW*(MW_air - MW_fluid)/(MW_fluid*(Y-1) - MW_air*Y))*dYdS
        Cp      = Y*(Cp_fluid - Cp_air) + Cp_air 
        dCpdS   = (Cp_fluid - Cp_air)*dYdS
        rhoh    = Pamb/const.R*MW*Cp
        drhohdS = Pamb/const.R*(MW*dCpdS + Cp*dMWdS)
        ##########################################################
        # TODO: integrating the energy equation without involving Cp - not sure what the isssue is in the code below
        # drhodS  = np.array([zero,                                                #d/dS(V_cl)
                            # -2*r**2*(rho_amb - rho_cl)*np.exp(r**2/(lam*B)**2),  #d/dS(B)
                            # lam**2*B**3*np.exp(r**2/(lam*B)**2),                 #d/dS(rho_cl)
                            # zero,                                                #d/dS(Y_cl)
                            # zero                                                 #d/dS(theta)
                            # ])*1./(lam**2*B**3)
        # # TODO: remove ideal gas assumption here (low priority)
        # T = Pamb*MW/(const.R*rho)
        # dTdS = Pamb/(const.R*rho)*dMWdS - Pamb*MW/(const.R*rho**2)*drhodS
        # h_amb = self._h_amb(T) #self.ambient.therm.h(T = T, P = Pamb)
        # d_h_amb_dT = self._dh_amb_dT(T)
        # h_fluid = self._h_fluid(T)
        # d_h_fluid_dT = self._dh_fluid_dT(T)
        # h = Y*h_fluid - Y*h_amb + h_amb
        # dhdS = (h_fluid - h_amb)*dYdS + Y*(d_h_fluid_dT - d_h_amb_dT)*dTdS + d_h_amb_dT*dTdS
        
        # rhoh = rho*h
        # drhohdS = h*drhodS + rho*dhdS
        # #########################################################
        
        
        # governing equations:
        LHScont = np.array([(lam**2*rho_cl + rho_amb)*B**2,                        #d/dS(V_cl)
                            2*(lam**2*rho_cl + rho_amb)*B*V_cl,                    #d/dS(B)
                            lam**2*B**2*V_cl,                                      #d/dS(rho_cl)
                            0,                                                    #d/dS(Y_cl)
                            0])*const.pi/(lam**2 + 1)                             #d/dS(theta)
        RHScont = rho_amb*E
        
        LHSxmom = np.array([(2*lam**2*rho_cl+rho_amb)*B**2*V_cl*np.cos(theta),     #d/dS(V_cl)
                            (2*lam**2*rho_cl+rho_amb)*B*V_cl**2*np.cos(theta),     #d/dS(B)
                            lam**2*B**2*V_cl**2*np.cos(theta),                     #d/dS(rho_cl)
                            0,                                                    #d/dS(Y_cl)
                            -(2*lam**2*rho_cl+rho_amb)*(B*V_cl)**2*np.sin(theta)/2 #d/dS(theta)
                            ])*const.pi/(2*lam**2+1)        
        RHSxmom = 0
        
        LHSymom = np.array([(2*lam**2*rho_cl+rho_amb)*B**2*V_cl*np.sin(theta),     #d/dS(V_cl)
                            (2*lam**2*rho_cl+rho_amb)*B*V_cl**2*np.sin(theta),     #d/dS(B)
                            lam**2*B**2*V_cl**2*np.sin(theta),                     #d/dS(rho_cl)
                            0,                                                    #d/dS(Y_cl)
                            (2*lam**2*rho_cl+rho_amb)*(B*V_cl)**2*np.cos(theta)/2  #d/dS(theta)
                            ])*const.pi/(2*lam**2+1)                                 
        RHSymom = -const.pi*lam**2*const.g*(rho_cl - rho_amb)*B**2
        
        LHSspec = np.array([B*Y_cl*rho_cl,                                         #d/dS(V_cl)
                            2*V_cl*Y_cl*rho_cl,                                    #d/dS(B)
                            B*V_cl*Y_cl,                                           #d/dS(rho_cl)
                            B*V_cl*rho_cl,                                         #d/dS(Y_cl)
                            0,                                                    #d/dS(theta)
                            ])*const.pi*lam**2*B/(lam**2 + 1)                
        RHSspec = Yamb*RHScont
        
        LHSener = 2*const.pi*integrate.trapz(V*drhohdS*r + rhoh*dVdS*r, r)
        LHSener += [const.pi/(6*lam**2 + 2)*(3*lam**2*rho_cl+rho_amb)*B**2*V_cl**2, #d/dS(V_cl)
                    const.pi/(9*lam**2 + 3)*(3*lam**2*rho_cl+rho_amb)*V_cl**3*B,    #d/dS(B)
                    const.pi/(6*lam**2 + 2)*lam**2*B**2*V_cl**3,                    #d/dS(rho_cl)
                    0,                                                             #d/dS(Y_cl)
                    0]                                                             #d/dS(theta)
        
        RHSener = h_amb0*RHScont
        
        LHS = np.array([LHScont,
                        LHSxmom,
                        LHSymom,
                        LHSspec,
                        LHSener
                        ])
        RHS = np.array([RHScont,
                        RHSxmom,
                        RHSymom,
                        RHSspec,
                        RHSener])
        
        dz = np.append(np.linalg.solve(LHS,RHS), np.array([np.cos(theta), np.sin(theta)]), axis = 0)
        
        return dz
    
    def reshape(self, enclosure, showPlot = False):
        '''
        reshapes the plume to turn upwards, should it hit the enclosure wall, 
        and crops it so it stops at the ceiling
        '''
        if np.any(self.x > enclosure.Xwall):
            iwall = np.argmax(self.x > enclosure.Xwall)
            for k in ['S', 'rho_cl', 'V_cl', 'Y_cl', 'B', 'theta', 'y', 'x']:
                self.__dict__[k] = np.append(np.append(self.__dict__[k][:iwall],
                                                       np.interp(enclosure.Xwall, self.x, self.__dict__[k])),
                                             self.__dict__[k][iwall:])
            self.y[iwall+1:] = self.y[iwall] + self.S[iwall+1:] - self.S[iwall]
            self.x[iwall:] = enclosure.Xwall
            self.theta[iwall:] = np.pi/2
        if np.any(self.y > enclosure.H):
            iceil = np.argmax(self.y > enclosure.H)
            for k in ['S', 'rho_cl', 'V_cl', 'Y_cl', 'B', 'theta', 'x', 'y']:
                np.append(self.__dict__[k][:iceil],
                          np.interp(enclosure.H, self.y, self.__dict__[k]))
        if showPlot == True:
            plt.plot(self.x,self.y)
        return self

    def m_flammable(self, X_lean=None, X_rich=None, Hmax=np.inf):
        '''
        Calculates the amount of mass in the plume
        that is within the flammability limits
        
        Parameters
        ----------
        plume : Plume class
            Class of plume results
        X_lean : float, optional
            Mole fraction lower flammability limit
            Default is None: will use default LFL for fuel
        X_rich : float, optional
            Mole fraction upper flammability limit
            Default is None: will use default UFL for fuel
        Hmax : float, optional
            Maximum height for integration
            Default is infinity (np.inf)
        
        Outputs
        -------
        mass : float
            Flammable mass in plume, up to height H (kg)
        '''
        MW_fluid = self.fluid.therm.MW
        MW_air = self.ambient.therm.MW
        if X_lean is None:
            fuel_props = FuelProperties(self.fluid.species)
            X_lean = fuel_props.LFL
        if X_rich is None:
            fuel_props = FuelProperties(self.fluid.species)
            X_rich = fuel_props.UFL
        Ylean = X_lean * MW_fluid / (X_lean * MW_fluid + (1 - X_lean) * MW_air)
        Yrich = X_rich * MW_fluid / (X_rich * MW_fluid + (1 - X_rich) * MW_air)

        # Trim the plume down to below Hmax:
        S = np.copy(self.S)
        Y_cl = np.copy(self.Y_cl)
        B = np.copy(self.B)
        rho_cl = np.copy(self.rho_cl)
        if np.all(self.y > Hmax) or np.all(self.Y_cl < Ylean) or np.all(self.Y_cl > Yrich): # no flammable mass 
            return 0
        elif np.any(self.y > Hmax): # trim arrays to Hmax
            S = np.append(S[np.argwhere(self.y < Hmax)].T[0], np.interp(Hmax, self.y, S))
            Y_cl = np.append(Y_cl[np.argwhere(self.y < Hmax)].T[0], np.interp(Hmax, self.y, Y_cl))
            B = np.append(B[np.argwhere(self.y < Hmax)].T[0], np.interp(Hmax, self.y, B))
            rho_cl = np.append(rho_cl[np.argwhere(self.y < Hmax)].T[0], np.interp(Hmax, self.y, rho_cl))
                
        def rhoY(r, i):
            'mass fraction and density at radius: r, for plume at node: i'
            rho = (rho_cl[i]-self.ambient.rho)*np.exp(-r**2/(self.lam*B[i]**2))+self.ambient.rho
            Y = rho_cl[i]*Y_cl[i]*np.exp(-r**2/(self.lam*B[i])**2)/rho
            return rho, Y

        Slean = np.interp(Ylean, Y_cl[::-1], S[::-1])
        Srich = np.interp(Yrich, Y_cl[::-1], S[::-1])
        ivals = slice(np.argmax(Y_cl <= Yrich), np.argmax(Y_cl <= Ylean))

        Y_cl, B, rho_cl, S = [np.append(np.append(np.interp(Srich, S, var), var[ivals]), np.interp(Slean, S, var)) for var in [Y_cl, B, rho_cl, S]]
        # radius of flammable concentration at each node:
        r_lean = np.array([0 if rhoY(0, i)[1] <= Ylean else 
                           optimize.brentq(lambda r: rhoY(r, i)[1] - Ylean, 0, 100*B[i])
                           for i in range(len(S))])
        r_rich = np.array([0 if rhoY(0, i)[1] <= Yrich else 
                           optimize.brentq(lambda r: rhoY(r, i)[1] - Yrich, 0, 100*B[i])
                           for i in range(len(S))])

        # integrate to find the mass/length at each node
        mass_per_len = np.array([integrate.quad(lambda r: np.prod(rhoY(r, i))*2*const.pi*r, 
                                 r_r, r_l)[0] for i, r_r, r_l in zip(range(len(S)), r_rich, r_lean)])
        # integrate to find the total mass
        return integrate.trapz(mass_per_len, S)
    
    @property
    def _contourdata(self):
        """

        Returns
        -------
        x : ndarray
            horizontal positions along jet (m)
        y : ndarray
            vertical positions along jet (m)
        X : ndarray
            mole fractions
        Y : ndarray
            mass fractions
        v : ndarray
            velocities
        T : ndarray
            temperatures
        """
        iS = np.arange(len(self.S))
        
        poshalf = np.logspace(-5, np.log10(3*np.max(self.B)))
        r = np.concatenate((-1 * poshalf[::-1], [0], poshalf))
        
        r, iS = np.meshgrid(r, iS)
        B = self.B[iS]
        rho_cl = self.rho_cl[iS]
        Y_cl = self.Y_cl[iS]
        V_cl = self.V_cl[iS]                    
        
        rho_amb, Tamb, Pamb = self.ambient.rho, self.ambient.T, self.ambient.P
        MW_fluid, MW_air = self.fluid.therm.MW, self.ambient.therm.MW
        
        rho = rho_amb + (rho_cl - rho_amb)*np.exp(-r**2/self.lam**2/B**2)
        Y   = Y_cl*rho_cl*np.exp(-(r**2)/((self.lam*B)**2))/rho
        MW  = MW_air*MW_fluid/(Y*(MW_air-MW_fluid) + MW_fluid)
        
        X = Y*MW/MW_fluid
        v = V_cl*np.exp(-(r**2)/(B**2))
        T = Pamb*MW/(const.R*rho)
        
        x = self.x[iS] + r*np.sin(self.theta[iS])
        y = self.y[iS] - r*np.cos(self.theta[iS])
        return x, y, X, Y, v, T
    
    def _radial_profile(self, distance, ind_var = 'Y', nB = 3):
        '''
        Returns radial profile at a certain distance along the jet

        Parameters
        ----------
        distance: float
            distance along jet to plot profile
        ind_var: string
            independent variable to plot, either mass fraction ('Y'), mole fraction ('X') or velocity ('v')
        nB: int
            number of halfwidths to plot

        Returns
        -------
        [r, ind_var]: radial profile of independent variable from -nB*B to nB*B
        '''
        B = np.interp(distance, self.S, self.B)
        rho_cl = np.interp(distance, self.S, self.rho_cl)
        Y_cl = np.interp(distance, self.S, self.Y_cl)
        V_cl = np.interp(distance, self.S, self.V_cl)

        r = np.logspace(-5, np.log10(nB*B))
        r = np.concatenate((-1*r[::-1], [0], r))

        rho_amb, Tamb, Pamb = self.ambient.rho, self.ambient.T, self.ambient.P
        MW_fluid, MW_air = self.fluid.therm.MW, self.ambient.therm.MW

        rho = rho_amb + (rho_cl - rho_amb)*np.exp(-r**2/self.lam**2/B**2)
        Y   = Y_cl*rho_cl*np.exp(-(r**2)/((self.lam*B)**2))/rho
        MW  = MW_air*MW_fluid/(Y*(MW_air-MW_fluid) + MW_fluid)

        X = Y*MW/MW_fluid
        v = V_cl*np.exp(-(r**2)/(B**2))
        T = Pamb*MW/(const.R*rho)
        if ind_var == 'Y':
            return [r, Y]
        if ind_var == 'X':
            return [r, X]
        if ind_var == 'v':
            return [r, v]
        if ind_var == 'T':
            return [r, T]

    def get_contour_data(self):
        return self._contourdata

    def plot_moleFrac_contour(self, mcolors='w', xlims=None, ylims=None, vlims=(0,0.1),
                              contour_levels=['LFL'], add_colorbar=True, plot_color=None,
                              plot_title=None, x_label=None, y_label=None,
                              fig_params={}, plot_params={}, subplot_params={}, fig=None, ax=None):
        '''
        Creates mole fraction contour plot

        Sends parameters to phys._plots plot_contour function and runs it.
        '''
        return plot_contour(data_type='mole', jet_or_flame=self,
                            mcolors=mcolors, xlims=xlims, ylims=ylims, vlims=vlims,
                            contour_levels=contour_levels, add_colorbar=add_colorbar, plot_color=plot_color,
                            plot_title=plot_title, x_label=x_label, y_label=y_label,
                            fig_params=fig_params, plot_params=plot_params,
                            subplot_params=subplot_params, ax=ax, fig=fig)

    def plot_massFrac_contour(self, mcolors='w', xlims=None, ylims=None, vlims=(0,1),
                              contour_levels=None, addColorBar=True, plot_color=None,
                              plot_title=None, x_label=None, y_label=None,
                              fig_params={}, plot_params={}, subplot_params={}, fig=None, ax=None):
        '''
        Creates mass fraction contour plot

        Sends parameters to phys._plots plot_contour function and runs it.
        '''
        return plot_contour(data_type='mass', jet_or_flame=self,
                            mcolors=mcolors, xlims=xlims, ylims=ylims, vlims=vlims,
                            contour_levels=contour_levels, add_colorbar=addColorBar, plot_color=plot_color,
                            plot_title=plot_title, x_label=x_label, y_label=y_label,
                            fig_params=fig_params, plot_params=plot_params,
                            subplot_params=subplot_params, ax=ax, fig=fig)

    def plot_velocity_contour(self, mcolors='w', xlims=None, ylims=None, vlims=(0,0.1),
                              contour_levels=None, addColorBar=True, plot_color=None,
                              plot_title=None, x_label=None, y_label=None,
                              fig_params={}, plot_params={}, subplot_params={}, fig=None, ax=None):
        '''
        Creates velocity contour plot

        Sends parameters to phys._plots plot_contour function and runs it.
        '''
        return plot_contour(data_type='velocity', jet_or_flame=self,
                            mcolors=mcolors, xlims=xlims, ylims=ylims, vlims=vlims,
                            contour_levels=contour_levels, add_colorbar=addColorBar, plot_color=plot_color,
                            plot_title=plot_title, x_label=x_label, y_label=y_label,
                            fig_params=fig_params, plot_params=plot_params,
                            subplot_params=subplot_params, ax=ax, fig=fig)

    def get_streamline_distances_to_mole_fractions(self, fracs):
        """
        Returns the streamline distance to a given mole fraction

        Parameters
        ----------
        fracs: int, float, list-like
            Mole fraction(s) of interest

        Returns
        -------
        streamline distances to centerline mole fraction of interest
        """
        return np.interp(fracs, self.X_cl[::-1], self.S[::-1])

    def get_xy_distances_to_mole_fractions(self, fracs):
        """
        Calculates the minimum and maximum distances
        along the x- and y-axes to mole fraction value(s) of interest

        Parameters
        ----------
        fracs: int, float, list-like
            Mole fraction(s) of interest

        Returns
        -------
        distances: dictionary
            Minimum and maximum distances of the form:
            {
                contour1: {[(min_x, max_x), (min_y, max_y)]},
                contour2: {...}
            }
        """
        fracs = [fracs] if type(fracs) in [int, float] else fracs
        contours = np.sort(fracs)
        x_values, y_values, molefrac_values, __, __, __ = self._contourdata
        cs = plt.contour(x_values, y_values, molefrac_values, levels=contours)
        distances = {}
        for contour, clc in zip(contours, cs.collections):
            for path in clc.get_paths():
                xs, ys = path.vertices.transpose()

                xmin, xmax = min(xs), max(xs)
                xmin = 0 if xmin > 0 and xmax > 0 else xmin
                xmax = 0 if xmin < 0 and xmax < 0 else xmax

                ymin, ymax = min(ys), max(ys)
                ymin = 0 if ymin > 0 and ymax > 0 else ymin
                ymax = 0 if ymin < 0 and ymax < 0 else ymax

                distances[contour] = [(xmin, xmax), (ymin, ymax)]
        return distances



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import os

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mplpatch

from ..phys import _flame, _jet, _unconfined_overpressure
from ..utilities import misc_utils


def calc_thermal_effects(amb_fluid, rel_fluid, rel_angle,
                         site_length, site_width,
                         orifices, rel_humid,
                         not_nozzle_model,
                         locations, developing_flows = None, chem = None,
                         create_plots=True, output_dir=None, verbose=False):
    """
    Calculates thermal effects for all positions in QRA

    Parameters
    ----------
    amb_fluid : Fluid object
        Ambient fluid

    rel_fluid : Fluid object
        Release fluid

    rel_angle : float
        angle of release (0 is horizontal) (radians)

    site_length : float
        Facility length (m)

    site_width : float
        Facility width (m)

    orifices : list of Orifice objects
        Leak orifices objects to perform calculations on
        One for each leak size

    rel_humid : float
        Relative humidity between 0 and 1

    not_nozzle_model : {'yuce', 'ewan', 'birc', 'bir2', 'molk'}
        Notional nozzle model identifier (i.e. for under-expanded jet zone)

    locations : list of tuples
        List of locations at which to determine thermal effects,
        each location is a tuple of 3 coordinates (m):
        [(x1, y1, z1), (x2, y2, z2), ...]
        
    developing_flows : DevelopingFlow objects
        Specified DevelopingFlows - same length as orifices
        
    chem : Combustion object
        Specified Combustion chemistry

    create_plots : bool
        Whether plot files should be created

    output_dir : str
        file path to directory in which to create files and plots

    verbose : bool
        If True, extra output will be printed (default False)

    Returns : dict
    -------
        fluxes : ndarray
            [W/m2] Heat flux data for all positions and leaksizes
            1-D array, all leaksizes for location 1,
            then all leaksizes location 2, etc.

        all_pos_files : list of str
            position plot file paths
    """
    num_sizes = len(orifices)
    num_positions = len(locations)
    x_locations = [location[0] for location in locations]
    z_locations = [location[2] for location in locations]
    cons_momentum, notional_noz_t = misc_utils.convert_nozzle_model_to_params(not_nozzle_model, rel_fluid)

    all_qrads = np.zeros((num_sizes, num_positions))
    all_pos_filepaths = []
    if developing_flows is None:
        developing_flows = len(orifices)*[None]
    for i, (orifice, developing_flow) in enumerate(zip(orifices, developing_flows)):
        if developing_flow is None:
            flame = _flame.Flame(rel_fluid, orifice, amb_fluid,
                                 theta0=rel_angle,
                                 nn_conserve_momentum=cons_momentum, nn_T=notional_noz_t,
                                 developing_flow=developing_flow, chem = chem,
                                 verbose=verbose)
        else:
            flame = _flame.Flame.from_developed_flow(developing_flow=developing_flow, chem = chem,
                                                     verbose=verbose)

        fluxes = flame.generate_positional_flux(locations, rel_humid)

        # Each row is the heatflux for all locs for specific leak size
        all_qrads[i, :] = fluxes

        if create_plots:
            fluxes_kWm2 = fluxes / 1000
            fluxes_str = 'Radiative Heat Flux (kW/m$^2$)'
            now_str = misc_utils.get_now_str()
            orif_diam_mm = orifice.d * 1000
            pos_fname = 'HeatFluxPositionPlot{}_{}.png'.format(i, now_str)
            if output_dir is None:
                output_dir = misc_utils.get_temp_folder()
            plot_filepath = os.path.join(output_dir, pos_fname)
            pos_title = '{} mm Leak Size'.format(round(orif_diam_mm, 3))
            plot_effect_positions(fluxes_kWm2, fluxes_str,
                                  plot_filepath, pos_title,
                                  x_locations, z_locations,
                                  site_length, site_width)
            all_pos_filepaths.append(plot_filepath)

    # Flatten heatflux (all leaksize loc1, then all leaksize loc2, etc)
    # Corresponds to flattening by row which is C-style ordering
    qrads_flat = all_qrads.flatten(order='C')

    result_dict = {
        "fluxes": qrads_flat,
        "all_pos_files": all_pos_filepaths
    }

    return result_dict


def calc_overp_effects(orifices, notional_nozzle_model,
                       release_fluid, ambient_fluid, release_angle,
                       locations, site_length, site_width,
                       overp_method,
                       BST_mach_flame_speed=None, TNT_equivalence_factor=None, 
                       developing_flows = None,
                       create_plots=True, output_dir=None,
                       verbose=False):
    """
    Calculates overpressure effects for all positions in QRA

    Parameters
    ----------
    orifices : list of Orifice objects
        Leak orifices objects to perform calculations on
        One for each leak size

    notional_nozzle_model : {'yuce', 'ewan', 'birc', 'bir2', 'molk'}
        Notional nozzle model identifier (i.e. for under-expanded jet zone)

    release_fluid : Fluid object
        Release fluid

    ambient_fluid : Fluid object
        Ambient fluid

    release_angle : float
        Angle of release (0 is horizontal) (radians)

    locations : list of locations
        List of locations at which to determine overpressure effects,
        each location is a tuple of 3 coordinates (m):
        [(x1, y1, z1), (x2, y2, z2), ...]

    site_length : float
        Facility length (m)

    site_width : float
        Facility width (m)
    
    overp_method : {'bst', 'tnt', 'bauwens'}
        Overpressure harm model identifier

    developing_flows : DevelopingFlow objects
        Specified DevelopingFlow - same length as orifices

    BST_mach_flame_speed : float
        If overp_method is 'bst', this must be specified
        Mach flame speed:
        0.2, 0.35, 0.7, 1.0, 1.4, 2.0, 3.0, 4.0, 5.2
        Detonation should use 5.2

    TNT_equivalence_factor : float
        If overp_method is 'tnt', this must be specified
        TNT equivalency factor

    create_plots : bool
        Whether plot files should be created
        Default is True

    output_dir : str
        File path to directory in which to create files and plots
        Default is None, which will use a temporary directory
        in the current working directory

    verbose : bool
        If True, extra output will be printed (default is False)

    Returns : dict
    -------
        overpressures : ndarray
            [Pa] Peak overpressure results for each position and leak size
            1-D array, all leaksizes for location 1,
            then all leaksizes location 2, etc.

        impulses : ndarray
            [Pa*s] Impulse results for each position and leak size
            1-D array, all leaksizes for location 1,
            then all leaksizes location 2, etc.
            Note: impulses are np.nan if impulse does not exist
            for specified model

        all_pos_overp_files : list of str
            position plot file paths for peak overpressure by position

        all_pos_impulse_files : list of str
            position plot file paths for impulse by position
    """
    num_sizes = len(orifices)
    num_positions = len(locations)
    x_locations = [location[0] for location in locations]
    z_locations = [location[2] for location in locations]

    all_overpressures = np.zeros((num_sizes, num_positions))
    all_impulses = np.zeros((num_sizes, num_positions))
    all_pos_overp_filepaths = []
    all_pos_impulse_filepaths = []
    if developing_flows is None:
        developing_flows = len(orifices)*[None]
    for i, (orifice, developing_flow) in enumerate(zip(orifices, developing_flows)):
        nozzle_cons_momentum, notional_noz_t = misc_utils.convert_nozzle_model_to_params(notional_nozzle_model, release_fluid)
        if developing_flow is None:
            jet = _jet.Jet(release_fluid, orifice, ambient_fluid,
                        theta0=release_angle,
                        nn_conserve_momentum=nozzle_cons_momentum, nn_T=notional_noz_t, developing_flow=developing_flow, 
                        verbose=verbose)
        else:
            jet = _jet.Jet.from_developed_flow(developing_flow=developing_flow, verbose=verbose)

        method = overp_method.lower()
        if method == 'bst':
            over_pressure_model = _unconfined_overpressure.BST_method(jet_object=jet,
                                                                      mach_flame_speed=BST_mach_flame_speed)
        elif method == 'tnt':
            over_pressure_model = _unconfined_overpressure.TNT_method(jet_object=jet,
                                                                      equivalence_factor=TNT_equivalence_factor)
        elif method == 'bauwens':
            over_pressure_model = _unconfined_overpressure.Bauwens_method(jet_object=jet)
        else:
            raise ValueError('Invalid overpressure method name')

        overpressures = over_pressure_model.calc_overpressure(locations)
        impulses = over_pressure_model.calc_impulse(locations)
        all_overpressures[i, :] = overpressures
        all_impulses[i, :] = impulses

        if create_plots:
            overpressures_kPa = overpressures / 1000
            overpressures_str = 'Peak Overpressure (kPa)'
            now_str = misc_utils.get_now_str()
            orif_diam_mm = orifice.d * 1000  # mm
            pos_fname = 'OverpressurePositionPlot{}_{}.png'.format(i, now_str)
            if output_dir is None:
                output_dir = misc_utils.get_temp_folder()
            plot_filepath = os.path.join(output_dir, pos_fname)
            pos_title = '{} mm Leak Size'.format(round(orif_diam_mm, 3))
            plot_effect_positions(overpressures_kPa, overpressures_str,
                                  plot_filepath, pos_title,
                                  x_locations, z_locations,
                                  site_length, site_width)
            all_pos_overp_filepaths.append(plot_filepath)

            if not np.isnan(impulses).any():
                impulses_kPas = impulses / 1000
                impulses_str = 'Impulse (kPa*s)'
                now_str = misc_utils.get_now_str()
                pos_fname = 'ImpulsePositionPlot{}_{}.png'.format(i, now_str)
                plot_filepath = os.path.join(output_dir, pos_fname)
                plot_effect_positions(impulses_kPas, impulses_str,
                                      plot_filepath, pos_title,
                                      x_locations, z_locations,
                                      site_length, site_width)
                all_pos_impulse_filepaths.append(plot_filepath)


    # Flatten overpressures and impulses
    # (all leaksize loc1, then all leaksize loc2, etc)
    # Corresponds to flattening by row which is C-style ordering
    all_overpressures_flat = all_overpressures.flatten(order='C')
    all_impulses_flat = all_impulses.flatten(order='C')

    result_dict = {
        'overpressures': all_overpressures_flat, 
        'impulses': all_impulses_flat,
        'all_pos_overp_files': all_pos_overp_filepaths,
        'all_pos_impulse_files': all_pos_impulse_filepaths
    }
    return result_dict


def plot_effect_positions(effects, effect_label, filename, title,
                          x_locations, z_locations, length, width):
    """
    Plots positions on facility grid colored by effect
    (e.g., heat flux, peak overpressure, or impulse)

    Parameters
    ----------
    effects : array
        Effect values at each position for coloring points
    
    effect_label : string
        Text label for the effect being plotted

    filename : string
        filename for saving plot

    title : string
        title for plot

    x_locations : array
        x-dimensions of the locations to plot (m)

    z_locations : array
        z-dimensions of the locations to plot (m)

    length : float
        Facility length (m)
    """
    fig, ax = plt.subplots()
    
    # Blue square represents leak/release source
    ax.plot(0, 0, 'bs', markersize=8, label='Leak Source')
    
    # Plot facility border
    ax.add_patch(mplpatch.Rectangle((-.2, -.2), length+.2, width+.2, fill=None))
    
    # Arrow representing flame direction, does not represent length/width
    ax.arrow(0, 0, length/4, 0, head_width=0.7, alpha=1, head_length=0.7, linewidth=0.3, fc='blue', ec='blue')
    
    # Plot occupant positions, colored by effect
    effect_ln = ax.scatter(x_locations, z_locations, s=36, c=effects, linewidths=0.5, edgecolors='black', cmap = plt.get_cmap('plasma'))

    # Plot formatting, colorbar, and saving
    ax.set_aspect('equal') 
    for spine in ['top', 'bottom', 'left', 'right']:
        ax.spines[spine].set_visible(False)
    ax.set_xlabel('Length (m)')
    ax.set_ylabel('Width (m)')
    ax.set_title(title)
    cb = fig.colorbar(effect_ln)
    cb.set_label(effect_label)
    fig.savefig(filename, bbox_inches='tight')
    plt.close(fig)



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

from ._jet import Jet
from ._indoor_release import IndoorRelease
from ._flame import Flame
from ._comps import Fluid, Orifice, Source, Enclosure, Vent
from ._unconfined_overpressure import BST_method, TNT_method, Bauwens_method
from ._fuel_props import FuelProperties
from . import api



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

from . import analysis
from .component_failure import ComponentFailureSet, ComponentFailure
from .component_set import ComponentSet
from . import consequence
from . import distributions
from . import effects
from . import event_tree
from . import ignition_probs
from . import leaks
from . import leak_size_results
from . import pipe_size
from .positions import PositionGenerator
from . import probits
from . import risk


"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

from CoolProp.CoolProp import AbstractState
import numpy as np
from ..utilities import misc_utils


class FuelProperties:
    """
    Fuel Properties

    Requires any fuel with CoolProp chemical formula or 'other_fluid' in the _data dictionary below

    Currently Available fluids: 
        H2, CH4, C3H8, N2, CO2, CO, C2H6, C4H10, C5H12, C6H16

    Required Fields:
        LFL: Lower Flammability Limit, mole/volume fraction
        UFL: Upper Flammability Limit, mole/volume fraction
        dHc: Heat of Combustion, J/kg
        other_name: Full Name of Fuel, string
    """
    _data = {'H2': {'LFL': 0.04,
                    'UFL': 0.75,
                    'dHc': 1.20E+08,
                    'other_name': 'hydrogen'
                   },
             'CH4': {'LFL': 0.05,
                     'UFL': 0.15,
                     'dHc': 5.00E+07,
                     'other_name': 'methane'
                    },
             'C3H8': {'LFL': 0.021,
                      'UFL': 0.095,
                      'dHc': 4.64E+07,
                      'other_name': 'n-Propane'
                     },
             'N2': {'LFL': np.nan,
                    'UFL': np.nan,
                    'dHc': 0,
                    'other_name': 'nitrogen'
                    },
             'CarbonDioxide': {'LFL': np.nan,
                     'UFL': np.nan,
                     'dHc': 0,
                     'other_name': 'CO2'
                     },
             'CarbonMonoxide': {'LFL': 0.125,
                     'UFL': 0.74,
                     'dHc': 1.01e7,
                     'other_name': 'CO'
                     },
             'C2H6': {'LFL': 0.03,
                      'UFL': .124,
                      'dHc': 4.75e7,
                      'other_name': 'ethane'
                      }, 
             'n-Butane': {'LFL': 0.018,
                          'UFL': .084,
                          'dHc': 4.57e7,
                          'other_name': 'n-butane'
                          }, 
             'IsoButane': {'LFL': 0.018,
                           'UFL': .084,
                           'dHc': 4.57e7,
                           'other_name': 'isobutane'
                           }, 
             'n-Pentane': {'LFL': 0.014,
                           'UFL': .078,
                           'dHc': 4.51e7,
                           'other_name': 'n-pentane'
                           },
             'IsoPentane': {'LFL': 0.014,
                            'UFL': .078,
                            'dHc': 4.51e7,
                            'other_name': 'isopentane'
                            },
             'n-Hexane': {'LFL': 0.012,
                           'UFL': .074,
                           'dHc': 4.48e7,
                           'other_name': 'n-hexane'
                           },
            }

    def __init__(self, species):
        # if '[' not in species:
        if not misc_utils.is_parsed_blend_string(species):
            species = species.upper()
            if species not in list(self._data.keys()):
                if species.lower() not in [fuel['other_name'].lower() for fuel in self._data.values()]:
                    raise ValueError('unknown properties for %s' % species)
                else:
                    species = [k for k, v in self._data.items() if v['other_name'].lower() == species.lower()][0]
            self.species = species
            for k in ['LFL', 'UFL', 'dHc']:
                self.__dict__[k] = self._data[species][k]
            self.nC, self.nH, self.nO = self._count_atoms(species)
        else:
            self.species = species
            specnames = [s.split('[')[0] for s in species.split('&')]
            molefracs = [float(s.split('[')[1][:-1]) for s in species.split('&')]
            eos = AbstractState('HEOS', '&'.join(specnames))
            eos.set_mole_fractions(molefracs)
            massfracs = eos.get_mass_fractions()
            # heat of combustion and nC weighted by mass/mole fraction
            # LFL and UFL found using Le Chatelier's formula (https://doi.org/10.1016/j.jhazmat.2007.11.085)
            self.dHc = 0; self.nC, self.nH, self.nO = 0, 0, 0; invLFL = 0; invUFL = 0; XtotLFL = 0; XtotUFL = 0;
            for species, X, Y in zip(specnames, molefracs, massfracs):
                if species.upper() not in [fuel.upper() for fuel in self._data.keys()]:
                    if species.lower() not in [fuel['other_name'].lower() for fuel in self._data.values()]:
                        raise ValueError('unknown properties for %s' % species)
                    else:
                        species = [k for k, v in self._data.items() if v['other_name'].lower() == species.lower()][0]
                else: 
                    species = [k for k in self._data.keys() if k.lower() == species.lower()][0]
                self.dHc += self._data[species]['dHc']*Y
                count = FuelProperties._count_atoms(species)
                self.nC += count[0]*X; self.nH += count[1]*X; self.nO += count[2]*X
                if not np.isnan(self._data[species]['LFL']):
                    invLFL += X/self._data[species]['LFL']
                    XtotLFL += X
                if not np.isnan(self._data[species]['UFL']):
                    invUFL += X/self._data[species]['UFL']
                    XtotUFL += X
            self.LFL, self.UFL = XtotLFL/invLFL, XtotUFL/invUFL
                
    @staticmethod
    def _count_atoms(species):
        if   'hydrogen' in species.lower() or species == 'H2': nC, nH, nO = 0, 2, 0
        elif 'methane'  in species.lower() or species == 'CH4': nC, nH, nO = 1, 4, 0
        elif 'ethane'   in species.lower() or species == 'C2H6': nC, nH, nO = 2, 6, 0
        elif 'propane'  in species.lower() or species == 'C3H8': nC, nH, nO = 3, 8, 0
        elif 'butane'   in species.lower(): nC, nH, nO = 4, 10, 0
        elif 'pentane'  in species.lower(): nC, nH, nO = 5, 12, 0
        elif 'hexane'   in species.lower(): nC, nH, nO = 6, 14, 0
        elif 'carbondioxide' in species.lower(): nC, nH, nO = 1, 0, 2
        elif 'carbonmonoxide' in species.lower(): nC, nH, nO = 1, 0, 1
        elif 'nitrogen' in species.lower() or species == 'N2': nC, nH, nO = 0, 0, 0
        return nC, nH, nO
    
    def __str__(self):
        return f'{self.species}\n------------\nLFL: {self.LFL}\nUFL: {self.UFL}\nheat of combustion: {self.dHc*1e-6:.1f} MJ/kg'
        
    def __repr__(self):
        return self.__str__()



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import os
import datetime
import re

from CoolProp import CoolProp


def params_as_str(param_dict):
    """

    Parameters
    ----------
    param_dict : dict
        dict of parameter names and values

    Returns : str
    -------
    string of params and values, one per line

    """
    sorted_params = sorted(param_dict)
    msg = "PARAMETERS:"
    for param in sorted_params:
        msg += "\n{:<48s}{}: {}".format('', param, param_dict[param])
    return msg


def clean_name(name):
    """
    Convert string name to alphanumeric lower-case

    """
    parsed = re.sub(r'\W+', '', name.lower())
    return parsed


def parse_nozzle_model(name):
    """
    Determine correct nozzle model name by ensuring name string is lower-case alphanumeric (includes underscore).

    Parameters
    ----------
    name : str
        Name of notional nozzle model

    Returns
    -------

    """
    cleanstr = clean_name(name)
    if cleanstr in ['yuc', 'yuce', 'yuceil_otugen', 'yuceilotugen', 'yuceil-otugen']:
        ret_str = 'yuce'
    elif cleanstr in ['ewan', 'ewan_moodie', 'ewanmoodie', 'ewan-moodie']:
        ret_str = 'ewan'
    elif cleanstr in ['bir', 'birc', 'birch', 'birch1']:
        ret_str = 'birc'
    elif cleanstr in ['birch2' 'bir2', 'b2']:
        ret_str = 'bir2'
    elif cleanstr in ['molkov', 'molk', 'mol']:
        ret_str = 'molk'
    else:
        ret_str = cleanstr

    return ret_str


def convert_nozzle_model_to_params(nozzle_model, rel_fluid=None):
    """ Convert nozzle model to conservation and solve spec.
    Options are:

    Parameters
    ----------
    nozzle_model : {'yuce', 'ewan', 'birc', 'bir2', 'molk'}
        Notional nozzle model ID

    rel_fluid: Fluid or None
        Required for birch models, provides temperature

    Returns
    -------
    con_mom : bool
        whether model conserves mass and momentum (True) or mass only (False)

    t_param : str

    """
    cleaned = parse_nozzle_model(nozzle_model)
    if cleaned == 'yuce':
        con_mom = True
        t_param = 'solve_energy'

    elif cleaned == 'ewan':
        con_mom = False
        t_param = 'Tthroat'

    elif cleaned == 'birc':
        con_mom = False
        t_param = rel_fluid.T

    elif cleaned == 'bir2':
        con_mom = True
        t_param = rel_fluid.T

    elif cleaned == 'molk':
        con_mom = False
        t_param = 'solve_energy'

    else:  # includes hars
        raise ValueError("Nozzle model not convertible to parameters")

    return con_mom, t_param


def get_now_str():
    """ Generates a string-formatted time, down to seconds """
    now = datetime.datetime.now()
    return now.strftime('%Y%m%d-%H%M%S')


def parse_phase_key(key):
    """
    Convert phase string identifier into value for phys library.
    All values besides 'gas' and 'liquid' will be converted to None

    Parameters
    ----------
    key : str
        Phase identifier key

    Returns
    -------
    str or None

    """
    return key if key in ['gas', 'liquid'] else None


def is_valid_fluid_string(fluid_str):
    """ Validates fluid string based on CoolProp fluids. """
    valid = False
    for fluid in CoolProp.FluidsList():
        if fluid_str.lower() == str(fluid).lower():
            valid = True
            break

    return valid


def parse_blend_dict_into_coolprop_dict(blend_dict):
    """ Builds coolprop-ready dict with cleaned names from dict of fluids and concentrations.

    Parameters
    ----------
    blend_dict : dict
    fluids and concentrations, e.g. {'ch4': 0.96, 'h2': 0, 'n2': 0.04}

    """
    epsilon = 0.00001
    if abs(sum(blend_dict.values()) - 1) > epsilon:
        raise ValueError(f"Total blend mole fraction must equal 1.0. Provided fraction was {sum(blend_dict.values())}")
    species_dict = {}
    total_frac = 0
    for key, val in blend_dict.items():
        if val <= 0.:
            continue

        cleaned = parse_fluid_key(key)
        if is_valid_fluid_string(cleaned):
            val = round(val, 6)
            total_frac += val
            species_dict[cleaned] = val
        else:
            raise ValueError(f"'{cleaned}' is not a valid fluid name")

    return species_dict


def parse_coolprop_dict_into_string(cdict):
    """
    Converts dict of fluids and concentrations into concatenated string.

    Parameters
    ----------
    cdict : dict

    Returns
    -------

    """
    if len(list(cdict.values())) == 1:
        s = list(cdict.keys())[0]
    else:
        s = '&'.join(['%s[%f]' % (s, X) for s, X in zip(cdict.keys(), cdict.values())])
    return s


def is_parsed_blend_string(fluid_str):
    """
    True if string is combination of fuel names and concentrations.
    Examples: True for HYDROGEN[1.000] and [HYDROGEN[0.5]&METHANE[0.5]
    """
    return '[' in fluid_str or '&' in fluid_str


def parse_fluid_key(key: str) -> str:
    """
    Parses fluid name to match valid CoolProp fluid strings.
    Note that 'n-' fluid names must have a capitalized root when passed to CoolProp.
    i.e. n-propane will fail but n-Propane will work.

    """
    key = key.lower()
    if key in ['hydrogen', 'hy', 'h2']:
        result = 'Hydrogen'
    elif key in ['methane', 'ch4']:
        result = 'Methane'
    elif key in ['propane', 'c3h8', 'n-propane']:
        result = 'n-Propane'
    elif key in ['nitrogen', 'n2']:
        result = 'Nitrogen'
    elif key in ['carbondioxide', 'co2', 'carbon dioxide']:
        result = 'CarbonDioxide'
    elif key in ['carbonmonoxide', 'co', 'carbon monoxide']:
        result = 'CarbonMonoxide'
    elif key in ['ethane', 'c2h6']:
        result = 'Ethane'
    elif key in ['n-butane', 'nbutane', 'n-c4h10']:
        result = 'n-Butane'
    elif key in ['n-pentane', 'npentane', 'n-c5h12']:
        result = 'n-Pentane'
    elif key in ['n-hexane', 'nhexane', 'n-c6h14']:
        result = 'n-Hexane'

    else:
        matched_names = [fluid for fluid in CoolProp.FluidsList() if fluid.lower() == key]
        if matched_names:
            result = matched_names[0]
        else:
            raise ValueError(f'Name ({key}) not found in CoolProp')

    return result


def get_temp_folder(temp_dir_name='temp'):
    """
    Returns location of temporary folder
    and creates it if needed

    Parameters
    ----------
    temp_dir_name : str, optional
        Name of temporary folder (default is 'temp')

    Returns
    -------
    temp_dir_path : str
        absolute path to temporary folder
    """
    temp_dir_path = os.path.join(os.getcwd(), temp_dir_name)
    if not os.path.isdir(temp_dir_path):
        os.mkdir(temp_dir_path)
    return temp_dir_path



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import numpy as np

from . import distributions



class ComponentFailureSet:
    """

    Parameters
    ----------
    f_failure_override : float or None
        Manual frequency value for accidents/shutdown failure.
        Other parameters will be ignored if this is not None.

    num_vehicles : int
        Number of vehicles in use.

    daily_fuelings : int
        Number of fuelings per day for each vehicle

    vehicle_days : int
        Annual days of operations

    noz_po_dist : str
        ID of distribution for pop-off failure mode.

    noz_po_a : float
        Failure distribution parameter A

    noz_po_b : float or None
        Failure distribution parameter B. None if using Expected Value

    noz_ftc_dist : str
        ID of distribution for nozzle failure-to-close failure mode.

    noz_ftc_a : float
        Failure distribution parameter A

    noz_ftc_b : float or None
        Failure distribution parameter B. None if using Expected Value

    mvalve_ftc_dist : str
        ID of distribution for manual valve FTC failure mode.

    mvalve_ftc_a : float
        Failure distribution parameter A

    mvalve_ftc_b : float or None
        Failure distribution parameter B. None if using Expected Value

    svalve_ftc_dist : str
        ID of distribution for solenoid valve FTC failure mode.

    svalve_ftc_a : float
        Failure distribution parameter A

    svalve_ftc_b : float or None
        Failure distribution parameter B. None if using Expected Value

    svalve_ccf_dist : str
        ID of distribution for solenoid valve common-cause failure mode.

    svalve_ccf_a : float
        Failure distribution parameter A

    svalve_ccf_b : float or None
        Failure distribution parameter B. None if using Expected Value

    overp_dist : str
        ID of distribution for overpressure failure mode.

    overp_a : float
        Failure distribution parameter A

    overp_b : float or None
        Failure distribution parameter B. None if using Expected Value

    pvalve_fto_dist : str
        ID of distribution for pressure-relief valve failure-to-open failure mode.

    pvalve_fto_a : float
        Failure distribution parameter A

    pvalve_fto_b : float or None
        Failure distribution parameter B. None if using Expected Value

    driveoff_dist : str
        ID of distribution for driveoff failure mode.

    driveoff_a : float
        Failure distribution parameter A

    driveoff_b : float or None
        Failure distribution parameter B. None if using Expected Value

    coupling_ftc_dist : str
        ID of distribution for coupling failure-to-close failure mode.

    coupling_ftc_a : float
        Failure distribution parameter A

    coupling_ftc_b : float or None
        Failure distribution parameter B. None if using Expected Value

    """
    use_override = False
    f_failure = None
    noz_po = None
    noz_ftc = None
    mvalve_ftc = None
    svalve_ftc = None
    svalve_ccf = None
    overp = None
    pvalve_fto = None
    driveoff = None
    coupling_ftc = None

    p_overp_rupture = None
    f_overp_rupture = None
    p_driveoff = None
    f_driveoff = None
    p_nozzle_release = None
    f_nozzle_release = None
    p_sol_valves_ftc = None
    f_sol_valves_ftc = None
    p_mvalve_ftc = None
    f_mvalve_ftc = None

    def __init__(self, f_failure_override=None,
                 num_vehicles=20, daily_fuelings=2, vehicle_days=250,
                 noz_po_dist='beta', noz_po_a=0.5, noz_po_b=610415.5,
                 noz_ftc_dist='expv', noz_ftc_a=0.002, noz_ftc_b=None,
                 mvalve_ftc_dist='expv', mvalve_ftc_a=0.001, mvalve_ftc_b=None,
                 svalve_ftc_dist='expv', svalve_ftc_a=0.002, svalve_ftc_b=None,
                 svalve_ccf_dist='expv', svalve_ccf_a=0.000127659574468085, svalve_ccf_b=None,
                 overp_dist='beta', overp_a=3.5, overp_b=310289.5,
                 pvalve_fto_dist='logn', pvalve_fto_a=-11.7359368859313, pvalve_fto_b=0.667849415603714,
                 driveoff_dist='beta', driveoff_a=31.5, driveoff_b=610384.5,
                 coupling_ftc_dist='beta', coupling_ftc_a=0.5, coupling_ftc_b=5031, verbose=False):

        if f_failure_override is not None:
            # User provided vehicle fueling failure frequency directly so ignore individual events
            self.use_override = True
            self.f_failure = float(f_failure_override)

        else:
            # calculate events; will be applied to 100% release for now
            self.use_override = False
            self.noz_po = ComponentFailure('Nozzle', 'Pop-off', noz_po_dist, noz_po_a, noz_po_b)
            self.noz_ftc = ComponentFailure('Nozzle', 'Failure to close', noz_ftc_dist, noz_ftc_a, noz_ftc_b)
            self.mvalve_ftc = ComponentFailure('Manual valve', 'Failure to close',
                                               mvalve_ftc_dist, mvalve_ftc_a, mvalve_ftc_b)
            self.svalve_ftc = ComponentFailure('Solenoid valves', 'Failure to close',
                                               svalve_ftc_dist, svalve_ftc_a, svalve_ftc_b)
            self.svalve_ccf = ComponentFailure('Solenoid valves', 'Common-cause failure',
                                               svalve_ccf_dist, svalve_ccf_a, svalve_ccf_b)

            self.overp = ComponentFailure('Overpressure during fueling', 'Accident', overp_dist, overp_a, overp_b)
            self.pvalve_fto = ComponentFailure('Pressure-relief valve', 'Failure to open',
                                               pvalve_fto_dist, pvalve_fto_a, pvalve_fto_b)
            self.driveoff = ComponentFailure('Driveoff', 'Accident', driveoff_dist, driveoff_a, driveoff_b)
            self.coupling_ftc = ComponentFailure('Breakaway coupling', 'Failure to close',
                                                 coupling_ftc_dist, coupling_ftc_a, coupling_ftc_b)

            num_fuelings = num_vehicles * daily_fuelings * vehicle_days
            self.p_driveoff = np.around(self.driveoff.mean * self.coupling_ftc.mean, 20)
            self.f_driveoff = np.around(num_fuelings * self.p_driveoff, 20)

            self.p_overp_rupture = np.around(self.overp.mean * self.pvalve_fto.mean, 20)
            self.f_overp_rupture = np.around(num_fuelings * self.p_overp_rupture, 20)
            f_accidents = self.f_overp_rupture + self.f_driveoff

            self.p_nozzle_release = np.around(self.noz_po.mean + self.noz_ftc.mean, 20)
            self.f_nozzle_release = np.around(num_fuelings * self.p_nozzle_release, 20)

            self.p_sol_valves_ftc = np.around(self.svalve_ftc.mean ** 3. + self.svalve_ccf.mean, 20)
            self.f_sol_valves_ftc = np.around(num_fuelings * self.p_sol_valves_ftc, 20)

            self.p_mvalve_ftc = np.around(self.mvalve_ftc.mean, 20)
            self.f_mvalve_ftc = np.around(num_fuelings * self.p_mvalve_ftc, 20)

            # Combined
            p_shutdown_fail = self.p_sol_valves_ftc * self.mvalve_ftc.mean * self.p_nozzle_release
            f_shutdown_fail = num_fuelings * p_shutdown_fail
            self.f_failure = np.around(float(f_accidents + f_shutdown_fail), 20)

            if verbose:
                print("COMPONENT FAILURE DATA")
                print(f'{num_fuelings} fuelings ({num_vehicles} vehicles, {daily_fuelings} fuel/d, {vehicle_days} days)')
                print("Nozzle popoff {}, {}, {}".format(noz_po_dist, noz_po_a, noz_po_b))
                print("Nozzle FTC {}, {}, {}".format(noz_ftc_dist, noz_ftc_a, noz_ftc_b))
                print("MValve FTC {}, {}, {}".format(mvalve_ftc_dist, mvalve_ftc_a, mvalve_ftc_b))
                print("SValve FTC {}, {}, {}".format(svalve_ftc_dist, svalve_ftc_a, svalve_ftc_b))
                print("SValve CCF {}, {}, {}".format(svalve_ccf_dist, svalve_ccf_a, svalve_ccf_b))
                print("PR Valve FTO {}, {}, {}".format(pvalve_fto_dist, pvalve_fto_a, pvalve_fto_b))
                print("BreakCoup FTC {}, {}, {}".format(coupling_ftc_dist, coupling_ftc_a, coupling_ftc_b))
                print("Fueling overp {}, {}, {}".format(overp_dist, overp_a, overp_b))
                print("Driveoff {}, {}, {}".format(driveoff_dist, driveoff_a, driveoff_b))

                print("Driveoff P {:.3g}, F {:.3g}".format(self.p_driveoff, self.f_driveoff))
                print("Overpressure rupture P {:.3g}, F {:.3g}".format(self.p_overp_rupture, self.f_overp_rupture))
                print("Nozzle release P {:.3g}, F {:.3g}".format(self.p_nozzle_release, self.f_nozzle_release))
                print("Sol valve FTC P {:.3g}, F {:.3g}".format(self.p_sol_valves_ftc, self.f_sol_valves_ftc))
                print("Shutdown fail P {:.3g}, F {:.3g}".format(p_shutdown_fail, f_shutdown_fail))

        if verbose:
            print("Frequency of other failures: {:.3g}".format(self.f_failure))

    def __str__(self):
        if self.use_override:
            # User provided vehicle fueling failure frequency directly so ignore individual events
            return f"Fuel failure OVERRIDE: {self.f_failure}"

        else:
            return f"Fuel failure: {self.f_failure}"


class ComponentFailure:
    """

    Parameters
    ----------
    component : str
        Name of component

    mode : str
        Description of failure mode

    distr_type : str
        Name referencing distribution type. lognormal, beta, ev, normal, uniform

    a : float
        First parameter describing distribution. Depends on type.
        e.g. if lognormal, assume this is sigma

    b : float
        Second parameter describing distribution. Depends on type.
        e.g. if lognorm, assume this is mu
    """

    def __init__(self, component, mode, distr_type, a, b=None):
        if not distributions.has_distribution(distr_type):
            msg = "Component failure distribution key {} not recognized".format(distr_type)
            raise ValueError(msg)
        if a is None and b is None:
            msg = "Distribution parameters cannot both be None"
            raise ValueError(msg)

        self.component = component
        self.mode = self.failure_mode = mode
        self.distr_type = distr_type
        self.a = a
        self.b = b

        distr_class = distributions.get_distribution_class(self.distr_type)
        self.distr = distr_class(a, b)
        self.mean = self.p = self.distr.mean

    def __str__(self):
        return "Component failure: {} {} | {}, mean {:.3g}".format(self.component, self.mode, self.distr, self.mean)



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import warnings

import numpy as np
from scipy import integrate
import scipy.constants as const

from ._jet import Jet

########################################################################
# TODO: untested for alternate fuels
########################################################################

class LayeringJet(Jet):
    def layer_accumulation(self, t, Vol_conc0, enclosure, tol = 1e-4):
        '''
        Integrates the Lowesmith layer accumulation over time t

        Inputs:
        -------
        t : ndarray
            array of time values to integrate over
        Vol_conc: ndarray
            gas volume (m^3) and concentration (mol/m^3) at the first time in the t array (t[0])

        Returns:
        --------
        tuple of time, volume and concentration

        - note: just outputting final value - could change to output full array if desired
        '''
        
        r = integrate.ode(self._gov_eqns).set_integrator('dopri5', atol = tol, rtol = tol)
        
        r.set_f_params(enclosure)
        r.set_initial_value(Vol_conc0, t[0])

        T, Vc = [], []
        def solout(t, y):
            T.append(t)
            y[0] = max(0, min(y[0], enclosure.V))
            y[1] = max(0, min(y[1], 1))
            Vc.append(np.array(y))
        r.set_solout(solout)

        for tval in t[1:]:
            r.integrate(tval)
            if not r.successful():
                warnings.warn('break at %f sec'%r.t)
                break
        Vol, c = np.array(Vc).T
        t = np.array(T)
        return t, Vol, c

    def _gov_eqns(self, t, Vol_conc, enclosure):
        '''
        Returns the derivitive of the volume and concentration with respect to time
        for a buoyant gas from a released jet within a vented enclosure.  The
        model is directly from Lowesmith et al. IJHE, Vol 34, 2009.

        Parameters
        ----------
        t:  time
        Vol_conc:  array of [volume, concentration]
        enclosure: class describing enclosure

        Returns
        -------
        array of differential values [d(volume)/dt, d(concentration)/dt]
        '''
        Vol, c = Vol_conc  # Layer Volume, Layer Volume fraction (or mole fraction)

        H_layer = min(Vol/enclosure.A, enclosure.H)  # height of a uniform layer with volume Vol
        H_vent  = H_layer - (enclosure.H - enclosure.ceiling_vent.H) # amount of layer height being exhausted by ceiling vent
        y_layer = enclosure.H - H_layer  # y-coordinate of bottom of flammable layer

        B, v = np.interp(y_layer, self.y, self.B), np.interp(y_layer, self.y, self.V_cl)

        Qj      = const.pi * B**2 * v  # volumetric flow rate of jet into layer (m**3/s)
        Qs       = self.Q_jet # this is set in the IndoorRelease class - volumetric flow of jet into enclosure

        Qw      = enclosure.floor_vent.Qw  # volumetric flow rate into enclosure due to wind (m**3/s)
        if H_vent > 0: # layer is being exhausted
            # volumetric flowrate due to buoyancy
            QB = (enclosure.ceiling_vent.Cd * enclosure.ceiling_vent.A *
                  np.sqrt(const.g * c * (1 - (self.fluid.therm.MW / self.ambient.therm.MW)) * H_vent))
            Qin = np.sqrt(QB**2 + Qw**2)
            dVdt = Qj - (Qin + Qs)
            if H_layer == enclosure.H:
                # Layer fills entire enclosure volume
                dVdt = min(0, dVdt) # dVdt <= 0
        else: # Layer is at minimum value
            QB = 0
            Qin = np.sqrt(QB**2 + Qw**2)
            dVdt = max(0, Qj - (Qin + Qs)) # dVdt >= 0
    #        dVdt = Qj - (Qin + Qs)
        
        if c >= 1:
            dcdt = min(0, (Qs - c*Qj) / Vol)
        elif c <= 0:
            dcdt = max(0, (Qs - c*Qj) / Vol)
        else:
            dcdt = (Qs - c*Qj) / Vol
        return np.array([dVdt, dcdt])
